#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import scipy as sp
from scipy.interpolate import interp1d

from fibermat import *
from fibermat import Mat, Mesh


################################################################################
# Degrees of Freedom
################################################################################

def displacement(u):
    """ Return nodal displacements."""
    return u[...]


def rotation(u):
    """ Return nodal rotations."""
    return np.zeros_like(u)


def force(F):
    """ Return nodal forces."""
    return F[...]


def torque(F):
    """ Return nodal torques."""
    return np.zeros_like(F)


################################################################################
# Mechanical model
################################################################################

def stiffness(mesh, **kwargs):
    """
    Assemble the quadratic system to be minimized.

    """
    # Optional
    if mesh is None:
        mesh = Mesh()

    assert Mesh.check(mesh)

    # Get mesh data
    mask = (mesh.index.values < mesh.beam.values)
    fiber = mesh.fiber[mask].values
    i = mesh.index[mask].values
    j = mesh.beam[mask].values

    # Get material data
    mat = mesh.flags.mat
    fiber = mat.loc[fiber]
    l = mesh.s.loc[j].values - mesh.s.loc[i].values

    # Timoshenko number : Ψ² = E / G * (h / l) ^ 2
    k0 = np.pi / 4 * fiber[[*"Gbh"]].prod(axis=1).values / l
    i *= 1
    j *= 1

    # Create stiffness data
    row = np.array([
        i, i,
        j, j,
    ]).ravel()
    col = np.array([
        i, j,
        i, j,
    ]).ravel()
    data = np.array([
         k0, -k0,
        -k0,  k0,
    ]).ravel()

    # Initialize 𝕂 matrix
    K = sp.sparse.coo_matrix((data, (row, col)),
                             shape=(1 * len(mesh), 1 * len(mesh)))

    # Initialize 𝒖 and 𝑭 vectors
    u = np.zeros(K.shape[0])
    F = np.zeros(K.shape[0])
    du = np.zeros(K.shape[0])
    dF = np.zeros(K.shape[0])

    return K, u, F, du, dF


def constraint(mesh, **kwargs):
    """
    Assemble the linear constraints.

    """
    # Optional
    if mesh is None:
        mesh = Mesh()

    assert Mesh.check(mesh)

    # Get mesh data
    mask = (mesh.index.values <= mesh.constraint.values)
    i = mesh.index[mask].values
    j = mesh.constraint[mask].values
    k = np.arange(len(i))
    O = i * 0  # : zero
    I = O + 1  # : one

    # Get material data
    mat = mesh.flags.mat
    mesh["h"] = mat.h.loc[mesh.fiber].values
    zi = mesh.z.loc[i].values
    zj = mesh.z.loc[j].values
    hi = mesh.h.loc[i].values
    hj = mesh.h.loc[j].values
    Z = (mesh.z.values + 0.5 * mesh.h.values).max()  # : upper boundary position
    i *= 1
    j *= 1
    k *= 3

    # Create constraint data
    row = np.array([k, k + 1, k + 1, k + 2]).ravel()
    col = np.array([i, i, j, j]).ravel()
    data = np.array([-I, I, -I, I]).ravel()

    # Initialize ℂ matrix
    C = sp.sparse.coo_matrix((data, (row, col)),
                             shape=(3 * len(mesh[mask]), 1 * len(mesh)))

    # Initialize 𝒇 and 𝑯 vectors
    f = np.zeros(C.shape[0])
    H = np.zeros(C.shape[0])
    df = np.zeros(C.shape[0])
    dH = np.zeros(C.shape[0])
    # (X₁ + u₁) ≥ ½h₁ ⟺ -u₁ ≤ X₁ - ½h₁
    H[::3] += zi - 0.5 * hi
    # (X₂ + u₂) - (X₁ + u₁) ≥ ½(h₁ + h₂) ⟺ u₁ - u₂ ≤ X₂ - X₁ - ½(h₁ + h₂)
    H[1::3] += zj - zi - 0.5 * (hi + hj)
    # (X₂ + u₂) ≤ Z - ½h₂ ⟺ u₂ ≤ Z - X₂ - ½h₂
    H[2::3] += Z - zj - 0.5 * hj
    dH[2::3] = 1
    # For end nodes
    H[1::3][mesh[mask].index == mesh[mask].constraint.values] = np.inf

    return C, f, H, df, dH


################################################################################
# Main
################################################################################

if __name__ == "__main__":

    # from fibermat import *

    # Generate a set of fibers
    mat = Mat(100)
    # Build the fiber network
    net = Net(mat)
    # Stack fibers
    stack = Stack(net)
    # Create the fiber mesh
    mesh = Mesh(stack)

    # Assemble the quadratic programming system
    K, u, F, du, dF = stiffness(mesh)
    C, f, H, df, dH = constraint(mesh)
    P = sp.sparse.bmat([[K, C.T], [C, None]], format='csc')
    # Permutation of indices
    perm = sp.sparse.csgraph.reverse_cuthill_mckee(P, symmetric_mode=True)
    # Enhanced solver
    spsolve = lambda A, b: sp.sparse.linalg.spsolve(A, b, use_umfpack=False)
    # # Visualize the system
    # fig, ax = plt.subplots(1, 2, figsize=(2 * 6.4, 4.8))
    # plot_system((K, u, F, du, dF), (C, f, H, df, dH), perm=None, ax=ax[0])
    # plot_system((K, u, F, du, dF), (C, f, H, df, dH), perm=perm, ax=ax[1])
    # plt.show()

    # Solve the mechanical packing problem
    K, C, u, f, F, H, Z, rlambda, mask, err = solve(
        mesh,
        stiffness(mesh),
        constraint(mesh),
        packing=4,
        solve=spsolve,
        perm=perm,
    )

    # # Visualize the system
    # fig, ax = plt.subplots(1, 2, figsize=(2 * 6.4, 4.8))
    # plot_system((K, u(0), F(0), du, dF), (C, f(0), H(0), df, dH), ax=ax[0])
    # plot_system((K, u(1), F(1), du, dF), (C, f(1), H(1), df, dH), ax=ax[1])
    # plt.show()

    # Export as VTK
    msh = vtk_mesh(
        mesh,
        displacement(u(1)),
        rotation(u(1)),
        force(f(1) @ C),
        torque(f(1) @ C),
    )
    msh.plot(scalars="force", cmap=plt.cm.twilight_shifted)
    # msh.save("outputs/msh.vtk")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pyvista as pv
from matplotlib import pyplot as plt
from scipy.interpolate import CubicHermiteSpline
from sklearn.neighbors import KDTree
from tqdm import tqdm

from fibermat import *
from fibermat import Mat, Mesh


def vtk_fiber(length=25., width=1., thickness=1., x=0., y=0., z=0.,
              u=1., v=0., w=0., shear=1., tensile=np.inf, index=None,
              r_resolution=1, theta_resolution=8, z_resolution=20, **kwargs):
    """
    Export a fiber as VTK mesh using `pyvista.CylinderStructured
    <https://docs.pyvista.org/version/stable/api/utilities/_autosummary/pyvista.CylinderStructured.html>`_.

    Parameters
    ----------
    length : float, optional
        Fiber length (mm). Default is 25 mm.
    width : float, optional
        Fiber width (mm). Default is 1 mm.
    thickness : float, optional
        Fiber thickness (mm). Default is 1 mm.
    x : float, optional
        Fiber position: X-coordinate (mm). Default is 0 mm.
    y : float, optional
        Fiber position: Y-coordinate (mm). Default is 0 mm.
    z : float, optional
        Fiber position: Z-coordinate (mm). Default is 0 mm.
    u : float, optional
        Fiber orientation: X-component. Default is 1.
    v : float, optional
        Fiber orientation: Y-component. Default is 0.
    w : float, optional
        Fiber orientation: Z-component. Default is 0.
    shear : float, optional
        Shear modulus (MPa). Default is 1 MPa.
    tensile : float, optional
        Tensile modulus (MPa). Default is ∞ MPa.
    index : int, optional
        Fiber label.

    Returns
    -------
    pyvista.StructuredGrid
        VTK mesh.

    Other Parameters
    ----------------
    r_resolution : int, optional
        Number of elements along the radius of the fiber. Default is 1.
    theta_resolution : int, optional
        Number of points on the circular face of the fiber. Default is 8.
    z_resolution : int, optional
        Number of points along the length of the fiber. Default is 20.
    kwargs :
        Additional keyword arguments ignored by the function.

    .. NOTE::
        If `index` is not None, the following fields are added to the VTK mesh:
            - `"fiber"` : fiber index
            - `"lbh"` : fiber dimensions (mm)
            - `"xyz"` : local fiber coordinates (mm)
            - `"uvw"` : fiber orientation vector
            - `"G"` : shear modulus (MPa)
            - `"E"` : tensile modulus (MPa)

    """
    # Create the VTK mesh (cylindrical structured grid)
    msh = pv.CylinderStructured(radius=np.linspace(0.5, 0, r_resolution + 1),
                                theta_resolution=theta_resolution,
                                z_resolution=z_resolution)

    l, b, h = length, width, thickness

    # Add fields to mesh data
    if index is not None:
        msh["fiber"] = np.full(len(msh.points), index)
        msh["lbh"] = np.tile([l, b, h], (len(msh.points), 1))
        msh["xyz"] = msh.points * np.array([[l, b, h]])
        msh["uvw"] = np.tile([u, v, w], (len(msh.points), 1))
        msh["G"] = np.full(len(msh.points), shear)
        msh["E"] = np.full(len(msh.points), tensile)

    # Transform the mesh (scale, rotate, and translate)
    msh.scale([l, b, h], inplace=True)
    pv.translate(msh,
                 center=(x, y, z),
                 direction=(u, v, w))

    # Return VTK mesh
    return msh


def vtk_mat(mat=None, func=None, verbose=True, **kwargs):
    """
    Export a :class:`~.Mat` object as VTK mesh.

    Parameters
    ----------
    mat : pandas.DataFrame, optional
        Set of fibers represented by a :class:`~.Mat` object.
    func : callable, optional
        Function called for each fiber to modify the mesh or add fields.
        It takes as arguments the VTK mesh and the label of the fiber.

    Returns
    -------
    pyvista.UnstructuredGrid
        VTK mesh.

    Other Parameters
    ----------------
    verbose : bool, optional
        If True, a progress bar is displayed. Default is True.
    kwargs :
        Additional keyword arguments passed to :meth:`vtk_fiber` function.

    .. NOTE::
        The following fields are added to the VTK mesh:
            - `"fiber"` : fiber index
            - `"lbh"` : fiber dimensions (mm)
            - `"xyz"` : local fiber coordinates (mm)
            - `"uvw"` : fiber orientation vector
            - `"G"` : shear modulus (MPa)
            - `"E"` : tensile modulus (MPa)

    """
    # Optional
    if mat is None:
        mat = Mat()

    assert Mat.check(mat)

    fibers = []  # : list to store individual fiber meshes

    for i in tqdm(mat.index, desc="Create VTK mat", disable=not verbose):
        # Get fiber
        fiber = mat.loc[i].astype(float)
        # Create the VTK mesh (cylindrical structured grid)
        msh = vtk_fiber(*fiber[[*"lbhxyzuvwGE"]].values,
                        index=i,
                        **kwargs)
        if func is not None:
            # Create additional fields
            func(msh, i)
        # Append fiber mesh to list
        fibers.append(msh)

    # Combine all individual fiber meshes into a single VTK mesh
    return pv.MultiBlock(fibers).combine()


def vtk_mesh(mesh=None,
             displacement=None, rotation=None,
             force=None, torque=None,
             verbose=True, **kwargs):
    """
    Export a :class:`~.Mesh` object as VTK mesh.

    Parameters
    ----------
    mesh : pandas.DataFrame, optional
        Fiber mesh represented by a :class:`~.Mesh` object.

    Returns
    -------
    pyvista.UnstructuredGrid
        VTK mesh.

    Other Parameters
    ----------------
    displacement : numpy.ndarray, optional
        Displacement field.
    rotation : numpy.ndarray, optional
        Rotation field.
    force : numpy.ndarray, optional
        Load field.
    torque : numpy.ndarray, optional
        Torque field.
    verbose : bool, optional
        If True, a progress bar is displayed. Default is True.
    kwargs :
        Additional keyword arguments passed to :meth:`vtk_fiber` function.

    .. HINT::
        The following fields are added to the VTK mesh:
            - `"fiber"` : fiber index
            - `"lbh"` : fiber dimensions (mm)
            - `"xyz"` : local fiber coordinates (mm)
            - `"uvw"` : fiber orientation vector
            - `"G"` : shear modulus (MPa)
            - `"E"` : tensile modulus (MPa)
        If `displacement` is not None:
            - `"displacement"` : displacement field (mm)
            - `"rotation"` : rotation field (rad)
            - `"curvature"` : curvature field (1 / mm)
        If `force` is not None:
            - `"force"` : force field (N)

    """
    # Optional
    if mesh is None:
        mesh = Mesh()

    assert Mesh.check(mesh)

    # Group nodes by fiber
    by_fiber = mesh.groupby("fiber")

    def interpolation_field(msh, i):
        # Prepare interpolation data
        fiber = by_fiber.get_group(i)
        s = fiber.s.values[:, None]
        x = msh["xyz"][:, [0]] * 0.9999
        k = KDTree(s).query(x, return_distance=False).ravel()
        s, x = s.ravel(), x.ravel()
        # Correct indices (s_k <= x_i < s_{k+1}, k \in [-1, n])
        k = k * (x > s[k]) + (k - 1) * (x <= s[k])
        # Indices containing relative distances (`np.floor(j) == k`)
        j = ((x - s[k]) / (s[k + 1] - s[k]))
        # Correct issues for periodic mesh
        j = np.array([*j])
        j[j == np.inf] = 0
        # Add to vtk_fiber
        j += fiber.index[k]
        msh["node"] = j

    # Create a VTK mesh with interpolation data
    mat = mesh.flags.mat
    msh = vtk_mat(mat, func=interpolation_field, verbose=verbose)

    if len(mat):
        # Interpolate fields
        s = np.arange(len(mesh))
        x = msh["node"]
        if displacement is not None:
            if rotation is None:
                rotation = 0 * displacement
            displacement = CubicHermiteSpline(s, displacement, rotation)
            msh["displacement"] = np.zeros(msh.points.shape)
            msh["displacement"][:, 2] = displacement(x)
            msh["rotation"] = displacement.derivative()(x)
            msh["curvature"] = displacement.derivative(2)(x)
            msh.points += msh["displacement"]
        if force is not None:
            if torque is None:
                torque = 0 * force
            force = CubicHermiteSpline(s, force, torque)
            msh["force"] = force(x)

    # Periodic boundary conditions (optional)
    if len(mat) and mesh.attrs["periodic"]:
        X = Y = mat.attrs["size"]
        Z1, Z2 = np.min(msh.points), np.max(msh.points)
        # Duplicate mesh for periodic conditions
        msh = pv.MultiBlock([
            msh,
            msh.copy().translate([-X, 0, 0]),
            msh.copy().translate([X, 0, 0]),
            msh.copy().translate([0, -Y, 0]),
            msh.copy().translate([0, Y, 0]),
            msh.copy().translate([-X, -Y, 0]),
            msh.copy().translate([-X, Y, 0]),
            msh.copy().translate([X, -Y, 0]),
            msh.copy().translate([X, Y, 0]),
        ]).combine().clip_box([-X, X, -Y, Y, Z1, Z2], invert=False)

    # Return VTK mesh
    return msh


################################################################################
# Main
################################################################################

if __name__ == "__main__":

    # from fibermat import *

    # Create a VTK fiber
    vtk_fiber().plot()

    # Generate a set of fibers
    mat = Mat(10)
    # Build the fiber network
    net = Net(mat, periodic=False)
    # Stack fibers
    stack = Stack(net)
    # Create the fiber mesh
    mesh = Mesh(stack)

    # Create a VTK mat
    vtk_mat(mat).plot()

    # Create a VTK mesh
    vtk_mesh(mesh).plot()

    # Solve the mechanical packing problem
    K, C, u, f, F, H, Z, rlambda, mask, err = solve(
        mesh,
        stiffness(mesh),
        constraint(mesh),
        packing=4,
    )

    # Export as VTK
    msh = vtk_mesh(
        mesh,
        displacement(u(1)),
        rotation(u(1)),
        force(f(1) @ C),
        torque(f(1) @ C),
    )
    msh.plot(scalars="force", cmap=plt.cm.twilight_shifted)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import scipy as sp
import warnings
from matplotlib import pyplot as plt
from scipy.interpolate import interp1d
from tqdm import tqdm

from fibermat import *
from fibermat import Mat


class Net(pd.DataFrame):
    r"""
    A class inherited from pandas.DataFrame_ to **build a fiber network**.

    It describes nodes and connections between fibers within a :class:`~.Mat` object:

        - *nodes* are defined as the nearest points between pairs of fibers.
        - *connections* link pairs of nodes to define relative positions between fibers.

    Parameters
    ----------
    mat : pandas.DataFrame, optional
        Set of fibers represented by a :class:`~.Mat` object.

    Other Parameters
    ----------------
    periodic : bool, optional
        If True, fibers are duplicated for periodicity. Default is True.
    pairs : numpy.ndarray, optional
        Pairs of fiber indices used to find nearest points. Size: (m x 2).
    kwargs :
        Additional keyword arguments ignored by the function.

    .. NOTE::
        The constructor calls :meth:`init` method if the object is instantiated with parameters.
        Otherwise, initialization is performed with the pandas.DataFrame_ constructor.

    .. _pandas.DataFrame: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html

    :Use:

        >>> # Generate a set of fibers
        >>> mat = Mat(100)
        >>> # Build the fiber network
        >>> net = Net(mat)
        >>> net
              A   B         sA         sB         xA         yA         zA         xB         yB         zB
        0     0   0  12.500000 -12.500000   6.057752  20.856058 -24.338157  -1.176401  -3.074404 -24.338157
        1     0   2   3.938063  -1.799582   3.580217  12.660413 -24.338157   3.580217  12.660413 -23.967450
        2     0   3   6.509881   8.253676   4.324414  15.122205 -24.338157   4.324414  15.122205 -23.766064
        3     0   5   0.269800  -7.165082   2.518746   9.149084 -24.338157   2.518746   9.149084 -21.802237
        4     0   6 -10.466114   6.264470  -0.587864  -1.127531 -24.338157  -0.587864  -1.127531 -21.637518
        ..   ..  ..        ...        ...        ...        ...        ...        ...        ...        ...
        862  95  95  12.500000 -12.500000 -14.234141  11.919304  23.096819 -17.446723 -12.873423  23.096819
        863  96  96  12.500000 -12.500000  15.557356 -19.115497  23.645974  -6.906063  -8.143040  23.645974
        864  97  97  12.500000 -12.500000 -12.739951 -17.721142  23.874757 -35.249295  -6.843209  23.874757
        865  98  98  12.500000 -12.500000  17.087939 -34.582099  24.091469  15.806064  -9.614985  24.091469
        866  99  99  12.500000 -12.500000 -17.894817 -13.721749  24.516947 -31.635635   7.163412  24.516947
        <BLANKLINE>
        [867 rows x 10 columns]

    Data
    ----
    + index : pandas.Index
        Connection label. Each label refers to a unique connection.
    + Pair of fibers:
        - A : pandas.Series
            First fiber label. It must satisfy `net.A` ≤ `net.B`.
        - B : pandas.Series
            Second fiber label. It must satisfy `net.A` ≤ `net.B`.
    + Curvilinear abscissa:
        - sA : pandas.Series
            Curvilinear abscissa of node along the first fiber (mm).
        - sB : pandas.Series
            Curvilinear abscissa of node along the second fiber (mm).
    + Relative node positions:
        - xA : pandas.Series
            X-coordinate of node along the first fiber (mm).
        - yA : pandas.Series
            Y-coordinate of node along the first fiber (mm).
        - zA : pandas.Series
            Z-coordinate of node along the first fiber (mm).
        - xB : pandas.Series
            X-coordinate of node along the second fiber (mm).
        - yB : pandas.Series
            Y-coordinate of node along the second fiber (mm).
        - zB : pandas.Series
            Z-coordinate of node along the second fiber (mm).

    ----

    Attributes
    ----------
    :attr:`attrs` :
        Global attributes of DataFrame.

    Methods
    -------
    :meth:`init` :
        Build a fiber network.
    :meth:`check` :
        Check that a :class:`Net` object is defined correctly.

    ----

    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the `Net` object.

        Parameters
        ----------
        args :
            Additional positional arguments passed to the constructor.
        kwargs :
            Additional keyword arguments passed to the constructor.

        See also
        --------
        `Net.init`.

        """
        if len(args) and Net._is(args[0]):
            # Initialize the DataFrame from argument
            super().__init__(*args, **kwargs)
            # Copy global attributes from argument
            self.attrs = args[0].attrs
            # Copy global flags from argument
            self.flags.mat = args[0].flags.mat

        else:
            # Initialize the DataFrame from parameters
            self.__init__(Net.init(*args, **kwargs))

        # Check `Net` object
        assert Net.check(self)

    # ~~~ Constructor ~~~ #

    @staticmethod
    def init(mat=None, periodic=True, pairs=None, **kwargs):
        """
        Build a fiber network.

        Parameters
        ----------
        mat : pandas.DataFrame, optional
            Set of fibers represented by a :class:`~.Mat` object.

        Returns
        -------
        net : pandas.DataFrame
            Initialized :class:`Net` object.

        Other Parameters
        ----------------
        periodic : bool, optional
            If True, fibers are duplicated for periodicity. Default is True.
        pairs : numpy.ndarray, optional
            Pairs of fiber indices used to find nearest points. Size: (m x 2).
        kwargs :
            Additional keyword arguments ignored by the function.

        """
        # Optional
        if mat is None:
            mat = Mat()

        assert Mat.check(mat)

        # Periodic boundary conditions (optional)
        if periodic:
            X = Y = mat.attrs["size"]
            # Duplicate fibers along boundaries
            l = mat.l.max()
            x1, x2 = -0.5 * X, 0.5 * X
            y1, y2 = -0.5 * Y, 0.5 * Y
            mask_x = mat.x < x1 + l
            mask_y = mat.y < y1 + l
            mask_xy = (mat.x > x2 - l) & (mat.y > y2 - l)
            # Left/right sides
            mat_x = mat[mask_x].copy()
            mat_x["inner"] = False
            mat_x["x"] += mat.attrs["size"]
            # Back/front sides
            mat_y = mat[mask_y].copy()
            mat_y["inner"] = False
            mat_y["y"] += mat.attrs["size"]
            # Corners
            mat_xy = mat[mask_xy].copy()
            mat_xy["inner"] = False
            mat_xy["x"] -= mat.attrs["size"]
            mat_xy["y"] -= mat.attrs["size"]
            # Merge repeated mats
            mat_ = pd.concat([mat, mat_x, mat_y, mat_xy])
        else:
            mat_ = mat

        # Pairwise indices
        if pairs is None:
            pairs = np.c_[np.triu_indices(len(mat_))]

        # Pairwise distances (optional)
        A = mat_[[*"xy"]].values[pairs][:, 0, :]
        B = mat_[[*"xy"]].values[pairs][:, 1, :]
        distances = np.linalg.norm(A - B, axis=1)
        # Spherical neighborhoods
        lengths = mat_[["l"]].values[pairs]
        diameters = np.mean(lengths, axis=1).ravel()
        pairs = pairs[distances < diameters]

        # Get material data
        L = mat_[[*"l"]].values[pairs]
        X = mat_[[*"xyz"]].values[pairs]
        U = mat_[[*"uvw"]].values[pairs]
        Xt = np.swapaxes(X, 1, 2)
        Ut = np.swapaxes(U, 1, 2)
        r = np.diff(Xt, axis=2)

        # Define normal projective system
        UUt = U @ Ut
        Ur = U @ r
        R = 0.5 * L
        mask = (np.linalg.det(UUt) != 0)
        # Solve normal equations: 𝕌ᵀ·R⃗ = r⃗ ⟺ 𝕌·𝕌ᵀ·R⃗ = 𝕌·r⃗
        R[mask] = np.linalg.solve(UUt[mask], Ur[mask])

        # Prepare node data
        pairs = mat_.index.values[pairs]
        abscissa = R * np.array([[1], [-1]])
        # points = X + abscissa * U

        # Remove nodes that do not belong to finite lines
        radius = 0.5 * mat_.l.values[pairs].reshape(-1, 2, 1)
        mask = np.all(np.abs(abscissa) < radius, axis=1).ravel()
        mask |= np.equal(*pairs.T)  # for fiber end points
        pairs = pairs[mask]
        # points = points[mask]
        abscissa = abscissa[mask]

        # Sort and remove repeated nodes (for periodicity)
        indices = np.argsort(pairs, axis=1)
        pairs = np.take_along_axis(pairs, indices, axis=1)
        abscissa = np.take_along_axis(abscissa[..., 0], indices, axis=1)
        df = (pd.DataFrame(np.c_[pairs, abscissa])
              .sort_values(by=[0, 1])
              .drop_duplicates([0, 1]))
        pairs = df[[0, 1]].values.astype(int)
        abscissa = df[[2, 3]].values.reshape(-1, 2, 1)
        points = (mat_[[*"xyz"]].values[pairs]
                  + abscissa * mat_[[*"uvw"]].values[pairs])

        # Initialize net DataFrame
        net = pd.DataFrame(
            data=np.c_[(pairs,
                        abscissa.reshape(-1, 2),
                        points.reshape(-1, 6))],
            columns=["A", "B", "sA", "sB", "xA", "yA", "zA", "xB", "yB", "zB"]
        )
        # Convert type to int
        net[[*"AB"]] = net[[*"AB"]].astype(int)

        # Set attributes
        net.attrs = mat.attrs
        net.attrs["periodic"] = periodic

        # Set flags
        net.flags.mat = mat

        # Return the `Net` object
        return net

    # ~~~ Public properties ~~~ #

    @property
    def attrs(self):
        """
        Global attributes of DataFrame:
            - n : int
                Number of fibers. By default, it is empty (n = 0).
            - size : float
                Box dimensions (mm). By default, the domain is a 50 mm square cube.
            - periodic : bool
                Boundary periodicity. By default, the domain is periodic.

        """
        return self._attrs

    @attrs.setter
    def attrs(self, attrs):
        self._attrs = attrs

    # ~~~ Public methods ~~~ #

    def check(self):
        """
        Check that a :class:`Net` object is defined correctly.

        This method is called when a :class:`Net` object is initialized.

        Raises
        ------
        KeyError
            If any keys are missing from the columns of :class:`Net` object.
        AttributeError
            If any attributes are missing from the dictionary :attr:`attrs`.
        IndexError
            If row indices are incorrectly defined:
                - Row indices are not unique in [0, ..., m-1] where m is the number of connections.
                - Connection labels are not sorted.
        TypeError
            If labels are not integers.
        ValueError
            If any of the following conditions are not met:
                - Fiber labels are incorrect.
                - There are duplicate connections.
                - Fiber labels are not ordered.

        Returns
        -------
        bool
            Indicates whether the object can be instantiated as :class:`Net`.

        .. TIP::
            - If `self` is None, it returns an empty :class:`Net` object.
            - If a `"skip_check"` flag is True in :attr:`attrs`, the check is passed.

        """
        if self is None:
            net = Net()
        else:
            net = self

        if "skip_check" in net.attrs.keys() and net.attrs["skip_check"]:
            warnings.warn("{}.attrs['skip_check'] is active."
                          " Delete it or set it to False.".format(net.__class__),
                          UserWarning)
            return True

        assert Mat.check(net.flags.mat)

        # Keys
        try:
            net[["A", "B", "sA", "sB", "xA", "yA", "zA", "xB", "yB", "zB"]]
        except KeyError as e:
            raise KeyError(e)

        # Attributes
        if not ("n" in net.attrs.keys()):
            raise AttributeError("'n' is not in attribute dictionary.")
        if not ("size" in net.attrs.keys()):
            raise AttributeError("'size' is not in attribute dictionary.")
        if not ("periodic" in net.attrs.keys()):
            raise AttributeError("'periodic' is not in attribute dictionary.")

        # Indices
        if not np.all(np.unique(net.index) == np.arange(len(net))):
            raise IndexError("Row indices must be unique in [0,..., {}]."
                             .format(len(net) - 1))
        if not np.all(net.index == np.arange(len(net))):
            raise IndexError("Connection labels must be sorted.")

        # Types
        if net[[*"AB"]].values.dtype != int:
            raise TypeError("Fiber labels are not integers.")

        # Data
        if len(net) and not (0 <= net[[*"AB"]].values.min()
                             and net[[*"AB"]].values.max() < net.attrs["n"]):
            raise ValueError("Fiber labels must be in [0,..., {}]."
                             .format(net.attrs["n"] - 1))
        if (len(net) != len(
                net[["A", "B", "sA", "sB"]].drop_duplicates(
                    ignore_index=True))):
            raise ValueError("Connections must be unique.")
        if not np.all(net.A <= net.B):
            raise ValueError("Pairs of fiber labels must be ordered: A ≤ B.")

        # Return True if the test is correct
        return True

    def _is(self):
        if self is None:
            return False
        try:
            __class__.check(self)
            return True
        except:
            return False


class Stack(Net):
    """
    A class inherited from :class:`Net` to **stack a set of fibers**.

    It solves the *linear programming system*:

    .. MATH::
        \min_{z} (-\mathbf{f} \cdot \mathbf{z})
        \quad s.t. \quad \mathbb{C} \, \mathbf{z} \leq \mathbf{H}
        \quad and \quad \mathbf{z} \geq \mathbf{h} / 2
    .. MATH::
        with \quad \mathbf{f} = -\mathbf{m} \, g \quad and \quad \mathbf{h} > 0

    where:
        - 𝐳 is the vector of fiber vertical positions (*unknowns of the problem*).
        - 𝐟 is the vector of fiber weights (with 𝐦 : fiber masses, 𝑔 gravity).
        - 𝐡 is the vector of fiber thicknesses.
        - ℂ is the matrix of inequality constraints that positions must satisfy to prevent the fibers from crossing each other.
        - -𝐇 corresponds to the minimum distances between the pairs of fibers.

    *Non-penetration conditions* between two fibers give the expressions of rows of ℂ and 𝐇:

    .. MATH::
        z_B - z_A \geq (h_A + h_B) \, / \, 2
        \quad \Leftrightarrow \quad z_A - z_B \leq - (h_A + h_B) \, / \, 2

    Parameters
    ----------
    net : pandas.DataFrame, optional
        Fiber network represented by a :class:`Net` object.

    Other Parameters
    ----------------
    threshold : float, optional
        Threshold distance value for proximity detection (mm).
    kwargs :
        Additional keyword arguments passed to the solver.

    .. NOTE::
        The constructor calls :meth:`init` method if the object is instantiated with parameters.
        Otherwise, initialization is performed with the pandas.DataFrame_ constructor.

    .. _pandas.DataFrame: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html

        :Use:

            >>> # Generate a set of fibers
            >>> mat = Mat(100)
            >>> # Build the fiber network
            >>> net = Net(mat)
            >>> # Stack fibers
            >>> stack = Stack(net)
            >>> stack
                  A   B         sA         sB         xA         yA    zA         xB         yB    zB
            0     0   0  12.500000 -12.500000   6.057752  20.856058   0.5  -1.176401  -3.074404   0.5
            1     0   2   3.938063  -1.799582   3.580217  12.660413   0.5   3.580217  12.660413   1.5
            2     0   3   6.509881   8.253676   4.324414  15.122205   0.5   4.324414  15.122205   2.5
            3     0   5   0.269800  -7.165082   2.518746   9.149084   0.5   2.518746   9.149084   1.5
            4     0   6 -10.466114   6.264470  -0.587864  -1.127531   0.5  -0.587864  -1.127531   1.5
            ..   ..  ..        ...        ...        ...        ...   ...        ...        ...   ...
            862  95  95  12.500000 -12.500000 -14.234141  11.919304  27.5 -17.446723 -12.873423  27.5
            863  96  96  12.500000 -12.500000  15.557356 -19.115497  27.5  -6.906063  -8.143040  27.5
            864  97  97  12.500000 -12.500000 -12.739951 -17.721142  27.5 -35.249295  -6.843209  27.5
            865  98  98  12.500000 -12.500000  17.087939 -34.582099  27.5  15.806064  -9.614985  27.5
            866  99  99  12.500000 -12.500000 -17.894817 -13.721749  26.5 -31.635635   7.163412  26.5
            <BLANKLINE>
            [867 rows x 10 columns]

    Data
    ----
    + index : pandas.Index
        Connection label. Each label refers to a unique connection.
    + Pair of fibers:
        - A : pandas.Series
            First fiber label. It must satisfy `net.A` ≤ `net.B`.
        - B : pandas.Series
            Second fiber label. It must satisfy `net.A` ≤ `net.B`.
    + Curvilinear abscissa:
        - sA : pandas.Series
            Curvilinear abscissa of node along the first fiber (mm).
        - sB : pandas.Series
            Curvilinear abscissa of node along the second fiber (mm).
    + Relative node positions:
        - xA : pandas.Series
            X-coordinate of node along the first fiber (mm).
        - yA : pandas.Series
            Y-coordinate of node along the first fiber (mm).
        - zA : pandas.Series
            Z-coordinate of node along the first fiber (mm).
        - xB : pandas.Series
            X-coordinate of node along the second fiber (mm).
        - yB : pandas.Series
            Y-coordinate of node along the second fiber (mm).
        - zB : pandas.Series
            Z-coordinate of node along the second fiber (mm).

    ----

    Attributes
    ----------
    :attr:`attrs` :
        Global attributes of DataFrame.

    Methods
    -------
    :meth:`init` :
        Stack fibers by gravity.
    :meth:`check` :
        Check that a :class:`Stack` object is defined correctly.
    :meth:`solve` :
        Solve the stacking problem.
    :meth:`constraint` :
        Assemble the linear system.

    ----

    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the `Stack` object.

        Parameters
        ----------
        args :
            Additional positional arguments passed to the constructor.
        kwargs :
            Additional keyword arguments passed to the constructor.

        See also
        --------
        `Stack.init`.

        """
        if len(args) and Stack._is(args[0]):
            # Initialize the DataFrame from argument
            super().__init__(*args, **kwargs)
            # Copy global flags from argument
            self.flags.mat = args[0].flags.mat

        else:
            # Initialize the DataFrame from parameters
            self.__init__(Stack.init(*args, **kwargs))

        # Check `Stack` object
        assert Stack.check(self)

    # ~~~ Constructor ~~~ #

    @staticmethod
    def init(net=None, threshold=None, **kwargs):
        """
        Stack fibers by gravity.

        Parameters
        ----------
        net : pandas.DataFrame, optional
            Fiber network represented by a :class:`Net` object.

        Returns
        -------
        stack : pandas.DataFrame
            Initialized :class:`Stack` object.

        Other Parameters
        ----------------
        threshold : float, optional
            Threshold distance value for proximity detection (mm).
        kwargs :
            Additional keyword arguments passed to the solver.

        .. WARNING::
            :class:`~.Mat` object is modified during execution.

        """
        # Optional
        if net is None:
            net = Net()

        assert Net.check(net)

        # Get material data
        mat = net.flags.mat

        # Solve the stacking problem
        linsol = Stack.solve(net, **kwargs)

        if linsol:
            # Update DataFrames
            mat.z = linsol.x
            net = Net(mat, **net.attrs)

            # Remove nodes based on threshold distances between nodes
            mask = np.zeros(len(net))
            # : 0 if removed node
            # : 1 if kept node
            # : 2 if end node
            if threshold is not None:
                mask[np.abs(net.zB.values - net.zA.values) <= threshold] = 1
                mask[net.A == net.B] = 2
                net = net[mask > 0]
                net.reset_index(drop=True, inplace=True)
                # Add global flags
                net.flags.mat = mat

        # Set attributes
        net.attrs["threshold"] = threshold

        # Initialize stack DataFrame
        stack = Stack(net)

        # Return the `Stack` object
        return stack

    # ~~~ Public properties ~~~ #

    @property
    def attrs(self):
        """
        Global attributes of DataFrame:
            - n : int
                Number of fibers. By default, it is empty (n = 0).
            - size : float
                Box dimensions (mm). By default, the domain is a 50 mm square cube.
            - periodic : bool
                Boundary periodicity. By default, the domain is periodic.
            - threshold : float
                Threshold distance value for proximity detection (mm).

        """
        return self._attrs

    @attrs.setter
    def attrs(self, attrs):
        self._attrs = attrs

    # ~~~ Public methods ~~~ #

    def check(self):
        """
        Check that a :class:`Stack` object is defined correctly.

        This method is called when a :class:`Stack` object is initialized.

        Raises
        ------
        KeyError
            If any keys are missing from the columns of :class:`Stack` object.
        AttributeError
            If any attributes are missing from the dictionary :attr:`attrs`.
        IndexError
            If row indices are incorrectly defined:
                - Row indices are not unique in [0, ..., m-1] where m is the number of connections.
                - Connection labels are not sorted.
        TypeError
            If labels are not integers.
        ValueError
            If any of the following conditions are not met:
                - Fiber labels are incorrect.
                - There are duplicate connections.
                - Fiber labels are not ordered.

        Returns
        -------
        bool
            Indicates whether the object can be instantiated as :class:`Stack`.

        .. TIP::
            - If `self` is None, it returns an empty :class:`Stack` object.
            - If a `"skip_check"` flag is True in :attr:`attrs`, the check is passed.

        """
        if self is None:
            stack = Stack()
        else:
            stack = self

        # Attributes
        if not ("threshold" in stack.attrs.keys()):
            raise AttributeError("'threshold' is not in attribute dictionary.")

        # Return True if the test is correct
        return Net.check(stack)

    def _is(self):
        if self is None:
            return False
        try:
            __class__.check(self)
            return True
        except:
            return False

    @staticmethod
    def solve(net=None, **kwargs):
        """
        Solve the stacking problem.

        Parameters
        ----------
        net : pandas.DataFrame, optional
            Fiber network represented by a :class:`Net` object.

        Returns
        -------
        linsol : OptimizeResult
            Results of linear programming solver.

        Other Parameters
        ----------------
        kwargs :
            Additional keyword arguments ignored by the function.

        .. SEEALSO::
            The solver is based on scipy.optimize.linprog_.

        .. _scipy.optimize.linprog: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.linprog.html

        """
        # Optional
        if net is None:
            net = Net()

        assert Net.check(net)

        # Assemble the linear programming system
        C, f, H, h = Stack.constraint(net)

        if net.attrs["n"]:
            # Linear programming solver
            bounds = np.c_[0.5 * h, np.full(len(h), np.inf)]
            # TODO: pass a solver with options as argument instead
            linsol = sp.optimize.linprog(f, C, H, bounds=bounds)
        else:
            linsol = None

        return linsol

    @staticmethod
    def constraint(net=None, **kwargs):
        """
        Assemble the linear system:

        .. MATH::
            \min_{z} (-\mathbf{f} \cdot \mathbf{z})
            \quad s.t. \quad \mathbb{C} \, \mathbf{z} \leq \mathbf{H}
            \quad and \quad \mathbf{z} \geq \mathbf{h} / 2
        .. MATH::
            with \quad \mathbf{f} = -\mathbf{m} \, g \quad and \quad \mathbf{h} > 0

        Parameters
        ----------
        net : pd.DataFrame, optional
            Fiber network represented by a :class:`Net` object.

        Returns
        -------
            C : sparse matrix
                Constraint matrix.
            mg : numpy.ndarray
                Force vector.
            H : numpy.ndarray
                Upper-bound vector.
            h : numpy.ndarray
                Thickness vector.

        Other Parameters
        ----------------
        kwargs :
            Additional keyword arguments ignored by the function.

        """
        # Optional
        if net is None:
            net = Net()

        assert Net.check(net)

        # Get network data
        mask = net.A.values < net.B.values
        i = net.A[mask].values
        j = net.B[mask].values
        k = 1 * np.arange(len(i))
        O = i * 0  # : zero
        I = O + 1  # : one

        # Get material data
        mat = net.flags.mat
        h = mat.h.values

        # Create constraint data
        row = np.array([k, k]).ravel()
        col = np.array([i, j]).ravel()
        data = np.array([I, -I]).ravel()

        # Initialize ℂ matrix
        C = sp.sparse.coo_matrix((data, (row, col)),
                                 shape=(1 * len(net[mask]), 1 * len(mat)))

        # Initialize 𝒇 and 𝑯 vectors
        mg = np.pi / 4 * mat[[*"lbh"]].prod(axis=1)  # : potential field
        H = np.zeros(C.shape[0])
        # X₂ - X₁ ≥ ½(h₁ + h₂) ⟺ X₁ - X₂ ≤ -½(h₁ + h₂)
        H -= 0.5 * (h[i] + h[j])

        return C, mg, H, h


################################################################################
# Main
################################################################################

if __name__ == "__main__":

    # from fibermat import *

    # Generate a set of fibers
    mat = Mat(10)
    # Build the fiber network
    net = Net(mat)
    # Stack fibers
    stack = Stack(net)

    # Get the linear system
    C, mg, H, h = Stack.constraint(stack)
    linsol = Stack.solve(stack)
    # Contact force
    f = linsol.ineqlin.marginals
    # Resulting force
    load = 0.5 * f @ np.abs(C) + 0.5 * f @ C

    # Check data
    Stack.check(stack)  # or `stack.check()`
    # -> returns True if correct, otherwise it raises an error.

    # Normalize by fiber weight
    load /= np.pi / 4 * mat[[*"lbh"]].prod(axis=1).mean()
    # Get loaded nodes
    points = (stack[stack.A < stack.B][["xA", "yA", "zA", "xB", "yB", "zB"]]
              .values.reshape(-1, 2, 3))
    # Prepare color scale
    cmap = plt.cm.viridis
    color = interp1d([np.min(load), np.max(load)], [0, 1])

    # Figure
    fig, ax = plt.subplots(subplot_kw=dict(projection='3d', aspect='equal',
                                           xlabel="X", ylabel="Y", zlabel="Z"))
    ax.view_init(azim=45, elev=30, roll=0)
    if len(mat):
        # Draw fibers
        for i in tqdm(range(len(mat)), desc="Draw fibers"):
            # Get fiber data
            fiber = mat.iloc[i]
            # Calculate fiber end points
            A = fiber[[*"xyz"]].values - 0.5 * fiber.l * fiber[[*"uvw"]].values
            B = fiber[[*"xyz"]].values + 0.5 * fiber.l * fiber[[*"uvw"]].values
            plt.plot(*np.c_[A, B], c=cmap(color(load[i])))
    if len(points):
        # Draw contacts
        for point in tqdm(points[~np.isclose(f, 0)], desc="Draw nodes"):
            plt.plot(*point.T, '--ok', lw=1, mfc='none', ms=3, alpha=0.2)
    # Set drawing box dimensions
    ax.set_xlim(-0.5 * stack.attrs["size"], 0.5 * stack.attrs["size"])
    ax.set_ylim(-0.5 * stack.attrs["size"], 0.5 * stack.attrs["size"])
    # Add a color bar
    norm = plt.Normalize(vmin=np.min(load), vmax=np.max(load))
    smap = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    cbar = plt.colorbar(smap, ax=ax)
    cbar.set_label("Load / $mg$ ($N\,/\,N$)")
    plt.show()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'FiberMat'
copyright = '2024, François Mahé'
author = 'François Mahé'
release = '1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.githubpages',
    'sphinx.ext.autosectionlabel',
    'myst_parser',
    'sphinx_copybutton',
]

templates_path = ['_templates']
exclude_patterns = []

# Disable some warnings (optional)
suppress_warnings = ['autosectionlabel.*', 'myst.xref_missing']

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = []



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from matplotlib import pyplot as plt

from fibermat import *


################################################################################
# Main
################################################################################

if __name__ == "__main__":

    # Generate a set of fibers
    mat = Mat(3000, length=4, width=0.2, thickness=0.05, size=40, tensile=1600)
    # Build the fiber network
    net = Net(mat, periodic=True)
    # Stack fibers
    stack = Stack(net, threshold=1)
    # Create the fiber mesh
    mesh = Mesh(stack)

    # Solve the mechanical packing problem
    K, C, u, f, F, H, Z, rlambda, mask, err = solve(
        mesh,
        stiffness(mesh),
        constraint(mesh),
        packing=4,
        itermax=10000,
    )

    # Export as VTK
    msh = vtk_mesh(
        mesh,
        displacement(u(1)),
        rotation(u(1)),
        force(f(1) @ C),
        torque(f(1) @ C),
    )
    msh.plot(scalars="force", cmap=plt.cm.twilight_shifted)
    # msh.save("outputs/msh.vtk")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import warnings
from matplotlib import pyplot as plt
from tqdm import tqdm

from fibermat import *


class Mat(pd.DataFrame):
    """
    A class inherited from pandas.DataFrame_ to **describe a fibrous material** made up of a set of random fibers.

    It defines:

        - the *geometry* of the straight fibers.
        - the *initial configuration* (positions and orientations).
        - the *material properties*.

    Parameters
    ----------
    n : int, optional
        Number of fibers. Default is 0.
    length : float, optional
        Fiber length (mm). Default is 25 mm.
    width : float, optional
        Fiber width (mm). Default is 1 mm.
    thickness : float, optional
        Fiber thickness (mm). Default is 1 mm.
    size : float, optional
        Box dimensions (mm). Default is 50 mm.
    theta : float, optional
        In plane angle (rad). Default is π rad.
    psi : float, optional
        Out-of-plane angle (rad). Default is 0 rad.
    shear : float, optional
        Shear modulus (MPa). Default is 1 MPa.
    tensile : float, optional
        Tensile modulus (MPa). Default is ∞ MPa.
    seed : int, optional
        Random seed for reproducibility. Default is 0.

    Other Parameters
    ----------------
    kwargs :
        Additional keyword arguments ignored by the function.

    .. NOTE::
        The constructor calls :meth:`init` method if the object is instantiated with parameters.
        Otherwise, initialization is performed with the pandas.DataFrame_ constructor.

    .. _pandas.DataFrame: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html

    :Use:

        >>> # Generate a set of fibers
        >>> mat = Mat(100)
        >>> mat
               l    b    h          x          y          z         u         v    w    G    E
        0   25.0  1.0  1.0   2.440675   8.890827 -24.338157  0.289366  0.957218 -0.0  1.0  inf
        1   25.0  1.0  1.0  10.759468 -11.499601 -24.178519  0.651721  0.758459  0.0  1.0  inf
        2   25.0  1.0  1.0   5.138169  11.759701 -23.967450  0.865730 -0.500512 -0.0  1.0  inf
        3   25.0  1.0  1.0   2.244159  23.109427 -23.766064  0.252040 -0.967717  0.0  1.0  inf
        4   25.0  1.0  1.0  -3.817260 -12.562343 -23.716864  0.957840 -0.287303  0.0  1.0  inf
        ..   ...  ...  ...        ...        ...        ...       ...       ...  ...  ...  ...
        95  25.0  1.0  1.0 -15.840432  -0.477060  23.096819  0.128503  0.991709 -0.0  1.0  inf
        96  25.0  1.0  1.0   4.325647 -13.629269  23.645974  0.898537 -0.438898 -0.0  1.0  inf
        97  25.0  1.0  1.0 -23.994623 -12.282176  23.874757  0.900374 -0.435117 -0.0  1.0  inf
        98  25.0  1.0  1.0  16.447001 -22.098542  24.091469  0.051275 -0.998685 -0.0  1.0  inf
        99  25.0  1.0  1.0 -24.765226  -3.279169  24.516947  0.549633 -0.835406  0.0  1.0  inf
        <BLANKLINE>
        [100 rows x 11 columns]

    Data
    ----
    + index : pandas.Index
        Fiber label. Each label refers to a unique fiber.
    + Fiber dimensions:
        - l : pandas.Series
            Fiber length (mm). By default, 25 mm long fibers are used.
        - b : pandas.Series
            Fiber width (mm). By default, 1 mm wide fibers are used.
        - h : pandas.Series
            Fiber thickness (mm). By default, 1 mm thick fibers are used.
    + Fiber position:
        - x : pandas.Series
            X-coordinate (mm). By default, positions are randomly distributed with a uniform distribution between the positions X: [-25 mm ; 25 mm].
        - y : pandas.Series
            Y-coordinate (mm). By default, positions are randomly distributed with a uniform distribution between the positions Y: [-25 mm ; 25 mm].
        - z : pandas.Series
            Z-coordinate (mm). By default, positions are randomly distributed with a uniform distribution between the positions Z: [-25 mm ; 25 mm].
    + Fiber orientation:
        - u : pandas.Series
            X-component. By default, orientations are randomly distributed with a uniform distribution in the half-unit circle θ: [-π / 2, π / 2].
        - v : pandas.Series
            Y-component. By default, orientations are randomly distributed with a uniform distribution in the half-unit circle θ: [-π / 2, π / 2].
        - w : pandas.Series
            Z-component. By default, orientations are in the plane, so the Z-component is 0.
    + Material properties:
        - G : pandas.Series
            Shear modulus (MPa). By default, shear modulus is 1 MPa.
        - E : pandas.Series
            Tensile modulus (MPa). By default, tensile modulus is infinite, which corresponds to fibers that are infinitely rigid in tension and bending.

    ----

    Attributes
    ----------
    :attr:`attrs` :
        Global attributes of DataFrame.

    Methods
    -------
    :meth:`init` :
        Generate a set of random straight fibers.
    :meth:`check` :
        Check that a :class:`Mat` object is defined correctly.

    ----

    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the `Mat` object.

        Parameters
        ----------
        args :
            Additional positional arguments passed to the constructor.

        Other Parameters
        ----------------
        kwargs :
            Additional keyword arguments passed to the constructor.

        See Also
        --------
        `Mat.init`.

        """
        if len(args) and Mat._is(args[0]):
            # Initialize the DataFrame from argument
            super().__init__(*args, **kwargs)
            # Copy global attributes from argument
            self.attrs = args[0].attrs

        else:
            # Initialize the DataFrame from parameters
            self.__init__(Mat.init(*args, **kwargs))

        # Check `Mat` object
        assert Mat.check(self)

    # ~~~ Constructor ~~~ #

    @staticmethod
    def init(n=0, length=25., width=1., thickness=1., size=50.,
             theta=np.pi, psi=0., shear=1., tensile=np.inf, seed=0, **kwargs):
        """
        Generate a set of random straight fibers.

        Parameters
        ----------
        n : int, optional
            Number of fibers. Default is 0.
        length : float, optional
            Fiber length (mm). Default is 25 mm.
        width : float, optional
            Fiber width (mm). Default is 1 mm.
        thickness : float, optional
            Fiber thickness (mm). Default is 1 mm.
        size : float, optional
            Box dimensions (mm). Default is 50 mm.
        theta : float, optional
            In plane angle (rad). Default is π rad.
        psi : float, optional
            Out-of-plane angle (rad). Default is 0 rad.
        shear : float, optional
            Shear modulus (MPa). Default is 1 MPa.
        tensile : float, optional
            Tensile modulus (MPa). Default is ∞ MPa.
        seed : int, optional
            Random seed for reproducibility. Default is 0.

        Returns
        -------
        mat : pandas.DataFrame
            Initialized :class:`Mat` object.

        Other Parameters
        ----------------
        kwargs :
            Additional keyword arguments ignored by the function.

        """
        # Random generation
        np.random.seed(seed)  # : random seed for reproducibility

        # Fiber dimensions
        l = np.full(n, length)  # : fiber length (mm)
        b = np.full(n, width)  # : fiber width (mm)
        h = np.full(n, thickness)  # : fiber thickness (mm)
        # Fiber position
        x = np.random.uniform(-0.5, 0.5, n) * size  # : X-coordinate (mm)
        y = np.random.uniform(-0.5, 0.5, n) * size  # : Y-coordinate (mm)
        z = np.random.uniform(-0.5, 0.5, n) * size  # : Z-coordinate (mm)
        z = np.sort(z)
        # Angles in radians
        p = np.random.uniform(-0.5, 0.5, n) * theta  # : in plane angle (rad)
        q = np.random.uniform(-0.5, 0.5, n) * psi  # : out-of-plane angle(rad)
        # Fiber orientation
        u = np.cos(p) * np.cos(q)  # : fiber orientation: X-component
        v = np.sin(p) * np.cos(q)  # : fiber orientation: Y-component
        w = np.sin(q)  # : fiber orientation: Z-component
        # Mechanical properties
        G = np.full(n, shear)  # : shear modulus (MPa)
        E = np.full(n, tensile)  # : tensile modulus (MPa)

        # Initialize mat DataFrame
        mat = pd.DataFrame(
            data=np.c_[l, b, h, x, y, z, u, v, w, G, E],
            columns=[*"lbhxyzuvwGE"]
        )

        # Set attributes
        mat.attrs["n"] = n
        mat.attrs["size"] = size

        # Return the `Mat` object
        return mat

    # ~~~ Public properties ~~~ #

    @property
    def attrs(self):
        """
        Global attributes of DataFrame:
            - n : int
                Number of fibers. By default, it is empty (n = 0).
            - size : float
                Box dimensions (mm). By default, the domain is a 50 mm square cube.

        """
        return self._attrs

    @attrs.setter
    def attrs(self, attrs):
        self._attrs = attrs

    # ~~~ Public methods ~~~ #

    def check(self):
        """
        Check that a :class:`Mat` object is defined correctly.

        This method is called when a :class:`Mat` object is initialized.

        Raises
        ------
        KeyError
            If any keys are missing from the columns of :class:`Mat` object.
        AttributeError
            If any attributes are missing from the dictionary :attr:`attrs`.
        IndexError
            If row indices are incorrectly defined:
                - Row indices are not unique in [0, ..., n-1] where n is the number of fibers.
                - Fiber labels are not sorted.
        ValueError
            If any of the following conditions are not met:
                - Dimensions are not positive.
                - Positions are not within a box of size specified in :attr:`attrs`.
                - Orientation vectors do not have unit lengths.
                - Material properties are not positive.

        Returns
        -------
        bool
            Indicates whether the object can be instantiated as :class:`Mat`.

        .. TIP::
            - If `self` is None, it returns an empty :class:`Mat` object.
            - If a `"skip_check"` flag is True in :attr:`attrs`, the check is passed.

        """
        if self is None:
            mat = Mat()
        else:
            mat = self

        if "skip_check" in mat.attrs.keys() and mat.attrs["skip_check"]:
            warnings.warn("{}.attrs['skip_check'] is active."
                          "Delete it or set it to False.".format(mat.__class__),
                          UserWarning)
            return True

        # Keys
        try:
            mat[[*"lbhxyzuvwGE"]]
        except KeyError as e:
            raise KeyError(e)

        # Attributes
        if not ("n" in mat.attrs.keys()):
            raise AttributeError("'n' is not in attribute dictionary.")
        if not ("size" in mat.attrs.keys()):
            raise AttributeError("'size' is not in attribute dictionary.")

        # Indices
        if len(mat) != mat.attrs["n"]:
            raise ValueError("Attribute `n: {}` does not correspond to"
                             " the number of fibers ({})."
                             .format(mat.attrs["n"], len(mat)))
        if (len(np.unique(mat.index)) != len(mat)
                or not np.all(np.unique(mat.index) == np.arange(len(mat)))):
            raise IndexError("Row indices must be unique in [0,..., {}]."
                             .format(len(mat) - 1))
        if not np.all(mat.index == np.arange(len(mat))):
            raise IndexError("Fiber labels must be sorted.")

        # Data
        if not np.all(mat[[*"lbh"]] > 0):
            raise ValueError("Dimensions must be positive.")
        if not np.all((-0.5 * mat.attrs["size"] <= mat[[*"xy"]])
                      & (mat[[*"xy"]] <= 0.5 * mat.attrs["size"])):
            raise ValueError("Positions must be in a box of size {}"
                             " (between {} and {})."
                             .format(mat.attrs["size"],
                                     -0.5 * mat.attrs["size"],
                                     0.5 * mat.attrs["size"]))
        if not np.allclose(np.linalg.norm(mat[[*"uvw"]], axis=1), 1):
            raise ValueError("Orientation vectors must have unit lengths.")
        if not np.all(mat[[*"GE"]] > 0):
            raise ValueError("Material properties must be positive.")

        # Return True if the test is correct
        return True

    def _is(self):
        if self is None:
            return False
        try:
            __class__.check(self)
            return True
        except:
            return False


################################################################################
# Main
################################################################################

if __name__ == "__main__":

    # from fibermat import *

    # Generate a set of fibers
    mat = Mat(10)

    # Get fiber data
    dimensions = mat[[*"lbh"]]  # size: (n x 3)
    positions = mat[[*"xyz"]]  # size: (n x 3)
    orientations = mat[[*"uvw"]]  # size: (n x 3)

    # Check data
    Mat.check(mat)  # or `mat.check()`
    # -> returns True if correct, otherwise it raises an error.

    # Figure
    fig, ax = plt.subplots(subplot_kw=dict(projection='3d', aspect='equal',
                                           xlabel="X", ylabel="Y", zlabel="Z"))
    ax.view_init(azim=45, elev=30, roll=0)
    if len(mat):
        # Draw fibers
        for i in tqdm(range(len(mat)), desc="Draw fibers"):
            # Get fiber data
            fiber = mat.iloc[i]
            # Calculate fiber end points
            A = fiber[[*"xyz"]].values - 0.5 * fiber.l * fiber[[*"uvw"]].values
            B = fiber[[*"xyz"]].values + 0.5 * fiber.l * fiber[[*"uvw"]].values
            plt.plot(*np.c_[A, B])
        # Set drawing box dimensions
        ax.set_xlim(-0.5 * mat.attrs["size"], 0.5 * mat.attrs["size"])
        ax.set_ylim(-0.5 * mat.attrs["size"], 0.5 * mat.attrs["size"])
    plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import ipywidgets
import numpy as np
from IPython.display import display, HTML

from fibermat import *


class Settings(dict):
    """
    A class for setting parameters in Jupyter notebooks.

    Methods
    -------
    :meth:`__iter__`
        Iterate over widgets.
    :meth:`__getitem__`
        Get the value of a given widget.
    :meth:`__setitem__`
        Set the value of a given widget.
    :meth:`__call__`
        Display widgets for setting parameters.

    ----

    """
    def __init__(self, n=0, length=25., width=1., thickness=1., size=50.,
                 theta=1., shear=1., tensile=np.inf, seed=0, periodic=True,
                 threshold=None, packing=1., itermax=1000, tol=1e-6, lmin=None,
                 lmax=None, coupling=1.0, interp_size=None, verbose=True):
        """
        Initialize a `Settings` widget.

        Parameters
        ----------
        TODO: Complete documentation.

        """
        # Input widgets
        self.n = ipywidgets.IntText(value=n, description="N")
        self.length = ipywidgets.FloatText(value=length, step=5.0, description="LENGTH")
        self.width = ipywidgets.FloatText(value=width, step=1.0, description="WIDTH")
        self.thickness = ipywidgets.FloatText(value=thickness, step=0.1, description="THICKNESS")
        self.size = ipywidgets.FloatText(value=size, step=10.0, description="SIZE")
        self.theta = ipywidgets.FloatSlider(value=theta, min=0.0, max=1.0, step=0.01, readout_format=".1%", description="THETA")
        self.shear = ipywidgets.FloatText(value=(shear if shear != np.inf else 1.0), step=1.0, layout=ipywidgets.Layout(width='100px'))
        self.shear.switch = ipywidgets.Checkbox(value=not (shear is np.inf), description="SHEAR", layout=ipywidgets.Layout(width='196px'))
        self.tensile = ipywidgets.FloatText(value=(tensile if tensile  != np.inf else shear * (length / thickness) ** 2), step=1.0, layout=ipywidgets.Layout(width='100px'))
        self.tensile.switch = ipywidgets.Checkbox(value=not (tensile is np.inf), description="TENSILE", layout=ipywidgets.Layout(width='196px'))
        self.seed = ipywidgets.IntText(value=(seed if seed is not None else 0), layout=ipywidgets.Layout(width='100px'))
        self.seed.switch = ipywidgets.Checkbox(value=not (seed is None), description="SEED", layout=ipywidgets.Layout(width='196px'))

        self.periodic = ipywidgets.Checkbox(value=periodic, description="PERIODIC")
        self.threshold = ipywidgets.FloatText(value=(threshold if threshold is not None else 20 * thickness), step=0.5, layout=ipywidgets.Layout(width='100px'))
        self.threshold.switch = ipywidgets.Checkbox(value=not (threshold is None), description="THRESHOLD", layout=ipywidgets.Layout(width='196px'))

        self.packing = ipywidgets.FloatText(value=packing, step=0.5, description="PACKING")
        self.itermax = ipywidgets.FloatLogSlider(value=itermax, min=0, max=6, step=1, readout_format=".0e", description="ITERMAX")
        self.tol = ipywidgets.FloatLogSlider(value=tol, min=-10.0, max=-1.0, step=1.0, readout_format=".0e", description="TOL")

        self.lmin = ipywidgets.FloatText(value=(lmin if lmin is not None else 0.01), step=0.01, layout=ipywidgets.Layout(width='100px'))
        self.lmin.switch = ipywidgets.Checkbox(value=not (lmin is None), description="LMIN", layout=ipywidgets.Layout(width='196px'))
        self.lmax = ipywidgets.FloatText(value=(lmax if lmax is not None else length), step=0.01, layout=ipywidgets.Layout(width='100px'))
        self.lmax.switch = ipywidgets.Checkbox(value=not (lmax is None), description="LMAX", layout=ipywidgets.Layout(width='196px'))
        self.coupling = ipywidgets.FloatText(value=coupling, step=0.01, description="COUPLING")

        self.interp_size = ipywidgets.IntText(value=(interp_size if interp_size is not None else 100), layout=ipywidgets.Layout(width='100px'))
        self.interp_size.switch = ipywidgets.Checkbox(value=not (interp_size is None), description="INTERPSIZE", layout=ipywidgets.Layout(width='196px'))
        self.verbose = ipywidgets.Checkbox(value=verbose, description="VERBOSE")

        def shear_on_changed(*args):
            self.shear.disabled = not self.shear.switch.value

        def tensile_on_changed(*args):
            self.tensile.disabled = not self.tensile.switch.value

        def seed_on_changed(*args):
            self.seed.disabled = not self.seed.switch.value

        def threshold_on_changed(*args):
            self.threshold.disabled = not self.threshold.switch.value

        def lmin_on_changed(*args):
            self.lmin.disabled = not self.lmin.switch.value

        def lmax_on_changed(*args):
            self.lmax.disabled = not self.lmax.switch.value

        def interp_size_on_changed(*args):
            self.interp_size.disabled = not self.interp_size.switch.value

        # Click events
        self.shear.switch.observe(shear_on_changed) ; shear_on_changed()
        self.tensile.switch.observe(tensile_on_changed) ; tensile_on_changed()
        self.seed.switch.observe(seed_on_changed) ; seed_on_changed()
        self.threshold.switch.observe(threshold_on_changed) ; threshold_on_changed()
        self.lmin.switch.observe(lmin_on_changed) ; lmin_on_changed()
        self.lmax.switch.observe(lmax_on_changed) ; lmax_on_changed()
        self.interp_size.switch.observe(interp_size_on_changed) ; interp_size_on_changed()

        # Initialize widget dictionary
        super().__init__(dict(
            n=self.n,
            length=self.length,
            width=self.width,
            thickness=self.thickness,
            size=self.size,
            theta=self.theta,
            shear=ipywidgets.HBox([self.shear.switch, self.shear]),
            tensile=ipywidgets.HBox([self.tensile.switch, self.tensile]),
            seed=ipywidgets.HBox([self.seed.switch, self.seed]),
            #
            periodic=self.periodic,
            threshold=ipywidgets.HBox([self.threshold.switch, self.threshold]),
            #
            packing=self.packing,
            itermax=self.itermax,
            tol=self.tol,
            #
            lmin=ipywidgets.HBox([self.lmin.switch, self.lmin]),
            lmax=ipywidgets.HBox([self.lmax.switch, self.lmax]),
            coupling=self.coupling,
            #
            interp_size=ipywidgets.HBox([self.interp_size.switch, self.interp_size]),
            verbose=self.verbose,
        ))

    # ~~~ Private methods ~~~ #

    def __iter__(self):
        """
        Iterate over widgets with syntax `*self`.

        Returns
        -------
        iterator
            Set of widgets as iterator.

        Notes
        -----
        Required for single star operator.

        """
        return iter(self.values())

    def __getitem__(self, name):
        """
        Get the value of a given widget with syntax `self[name]` and `**self`.

        Parameters
        ----------
        name : str
            Name of the widget.

        Notes
        -----
        Required for bracket accessor and double star operator.

        """
        if name == "theta":
            return self.__getattribute__(name).value * np.pi
        elif name in ["shear", "tensile"]:
            return (self.__getattribute__(name).value
                    if self.__getattribute__(name).switch.value
                    else np.inf)
        elif name in ["seed", "threshold", "lmin", "lmax", "interp_size"]:
            return (self.__getattribute__(name).value
                    if self.__getattribute__(name).switch.value
                    else None)
        elif name == "itermax":
            return int(self.__getattribute__(name).value)
        else:
            # Default getter
            return self.__getattribute__(name).value

    def __setitem__(self, name, value):
        """
        Set the value of a given widget with syntax `self[name] = value`.

        Parameters
        ----------
        name : str
            Name of the widget.
        value : Any
            Value of the widget.

        Notes
        -----
        Required for bracket accessor and setting via :meth:`__call__` method.

        """
        if name in ["shear", "tensile"]:
            self.__getattribute__(name).switch.value = (False if value is np.inf
                                                        else True)
            if value is not np.inf:
                self.__getattribute__(name).value = value
        elif name in ["seed", "threshold", "lmin", "lmax", "interp_size"]:
            self.__getattribute__(name).switch.value = (False if value is None
                                                        else True)
            if value is not None:
                self.__getattribute__(name).value = value
        else:
            # Default setter
            self.__getattribute__(name).value = value

    def __call__(self, **kwargs):
        """
        Display widgets for setting parameters with syntax `self(**kwargs)`.

        Other Parameters
        ----------------
        kwargs :
            Additional keyword arguments to set widget values.

        """
        for name, value in kwargs.items():
            self.__setitem__(name, value)
        if len(kwargs) == 0:
            display(*self)


################################################################################
# Main
################################################################################

if __name__ == "__main__":

    # from fibermat import *

    if "inputs" not in globals():
        inputs = Settings()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import scipy as sp
from scipy.interpolate import interp1d

from fibermat import *
from fibermat import Mat, Mesh


################################################################################
# Degrees of Freedom
################################################################################

def displacement(u):
    """ Return nodal displacements."""
    return u[..., ::2]


def rotation(u):
    """ Return nodal rotations."""
    return u[..., 1::2]


def force(F):
    """ Return nodal forces."""
    return F[..., ::2]


def torque(F):
    """ Return nodal torques."""
    return F[..., 1::2]


################################################################################
# Mechanical model
################################################################################

def stiffness(mesh, lmin=0.01, lmax=None, coupling=0.99, **kwargs):
    r"""
    Assemble the quadratic system to be minimized.

    The mechanical model is built using a **Timoshenko beam law** [1]_:

    .. MATH::
        \mathbb{K}_e = \frac{Gbh}{l_e} \cdot \frac{\pi / 4}{1 + \frac{G}{E} \left( \frac{l_e}{h} \right)^2}
            \left[\begin{matrix}
                1  &  l_e / 2  &  -1  &  l_e / 2  \\
                l_e / 2  &  {l_e}^2 / 3 + \frac{E}{G} h^2  &  -l_e / 2  &  {l_e}^2 / 6 - \frac{E}{G} h^2  \\
               -1  &  -l_e / 2  &  1  &  -l_e / 2  \\
                l_e / 2  &  {l_e}^2 / 6 - \frac{E}{G} h^2  &  -l_e / 2  &  {l_e}^2 / 3 + \frac{E}{G} h^2  \\
            \end{matrix}\right]
            \ , \quad \mathbf{F}_e =
            \left(\begin{matrix}
                0 \\
                0 \\
                0 \\
                0 \\
            \end{matrix}\right)

    where:
        - 𝑙ₑ is the length of the beam element.
        - 𝐸 is the tensile modulus.
        - 𝐺 is the shear modulus.
        - 𝑏 and h are the width and thickness of the fiber.

    The displacement vector :math:`\mathbf{u} = (\dots, u_i, \theta_i, \dots)`
    (with 𝑢ᵢ being the vertical displacement and θᵢ the rotation of the cross-section of the i-th node)
    satisfies *mechanical equilibrium*:

    .. MATH::
        \mathbb{K} \, \mathbf{u} = \mathbf{F}

    .. RUBRIC:: Footnotes

    .. [1] `Timoshenko–Ehrenfest beam theory, Wikipedia <https://en.wikipedia.org/wiki/Timoshenko%E2%80%93Ehrenfest_beam_theory>`_.

    Parameters
    ----------
    mesh : pandas.DataFrame
        Fiber mesh represented by a :class:`~.Mesh` object.

    Returns
    -------
    tuple
        K : sparse matrix
            Stiffness matrix (symmetric positive-semi definite).
        u : numpy.ndarray
            Displacement vector.
        F : numpy.ndarray
            Load vector.
        du : numpy.ndarray
            Incremental displacement vector.
        dF : numpy.ndarray
            Incremental load vector.

    Other Parameters
    ----------------
    lmin : float, optional
        Lower bound used to rescale beam lengths (mm). Default is 0.01 mm.
    lmax : float, optional
        Upper bound used to rescale beam lengths (mm).
    coupling : float, optional
        Coupling numerical constant between 0 and 1. Default is 0.99.
    kwargs :
        Additional keyword arguments ignored by the function.

    :Use:

        >>> from fibermat import *

        >>> # Linear model (Ψ² ≫ 1)
        >>> mat = Mat(1, length=1, width=1, thickness=1, shear=1, tensile=np.inf)
        >>> net = Net(mat)
        >>> mesh = Mesh(net)
        >>> # print("Linear (Ψ² ≫ 1) =")
        >>> print(4 / np.pi * stiffness(mesh, coupling=1)[0].todense())
        [[ 1.   0.5 -1.   0.5]
         [ 0.5  inf -0.5 -inf]
         [-1.  -0.5  1.  -0.5]
         [ 0.5 -inf -0.5  inf]]

        >>> # Timoshenko model (Ψ² = 1)
        >>> mat = Mat(1, length=1, width=1, thickness=1, shear=2, tensile=2)
        >>> net = Net(mat)
        >>> mesh = Mesh(net)
        >>> # print("Timoshenko (Ψ² = 1) = 1 / 2 *")
        >>> print(4 / np.pi * stiffness(mesh, coupling=1)[0].todense())
        [[ 1.          0.5        -1.          0.5       ]
         [ 0.5         1.33333333 -0.5        -0.83333333]
         [-1.         -0.5         1.         -0.5       ]
         [ 0.5        -0.83333333 -0.5         1.33333333]]

        >>> # Euler model (Ψ² ≪ 1)
        >>> mat = Mat(1, length=1, width=1, thickness=1, shear=1e12, tensile=12)
        >>> net = Net(mat)
        >>> mesh = Mesh(net)
        >>> # print("Euler (Ψ² ≪ 1) = 1 / 12 *")
        >>> print(4 / np.pi * stiffness(mesh, coupling=1)[0].todense())
        [[ 12.   6. -12.   6.]
         [  6.   4.  -6.   2.]
         [-12.  -6.  12.  -6.]
         [  6.   2.  -6.   4.]]

    """
    # Optional
    if mesh is None:
        mesh = Mesh()

    assert Mesh.check(mesh)

    # Get mesh data
    mask = (mesh.index.values < mesh.beam.values)
    fiber = mesh.fiber[mask].values
    i = mesh.index[mask].values
    j = mesh.beam[mask].values

    # Get material data
    mat = mesh.flags.mat
    fiber = mat.loc[fiber]
    l = mesh.s.loc[j].values - mesh.s.loc[i].values
    if lmin is None:
        lmin = np.min(l)
    if lmax is None:
        lmax = np.max(l)
    l = interp1d([min(np.min(l), lmin), max(np.max(l), lmax)], [lmin, lmax])(l)

    # Timoshenko number : Ψ² = E / G * (h / l) ^ 2
    k0 = np.pi / 4 * fiber[[*"Gbh"]].prod(axis=1).values / l
    k0 /= (1 + (fiber.G / fiber.E) * (l / fiber.h) ** 2)
    k1 = k0 * l / 2
    k1 *= coupling  # Numerical regularization
    k2 = k0 * l ** 2 / 3
    k2 += k0 * (fiber.E / fiber.G) * fiber.h ** 2
    k3 = k0 * l ** 2 / 2
    k4 = k2 - k3
    i *= 2
    j *= 2

    # Create stiffness data
    row = np.array([
        i + 0, i + 0, i + 0, i + 0,
        i + 1, i + 1, i + 1, i + 1,
        j + 0, j + 0, j + 0, j + 0,
        j + 1, j + 1, j + 1, j + 1,
    ]).ravel()
    col = np.array([
        i + 0, i + 1, j + 0, j + 1,
        i + 0, i + 1, j + 0, j + 1,
        i + 0, i + 1, j + 0, j + 1,
        i + 0, i + 1, j + 0, j + 1,
    ]).ravel()
    data = np.array([
         k0,  k1, -k0,  k1,
         k1,  k2, -k1, -k4,
        -k0, -k1,  k0, -k1,
         k1, -k4, -k1,  k2
    ]).ravel()

    # Initialize 𝕂 matrix
    K = sp.sparse.coo_matrix((data, (row, col)),
                             shape=(2 * len(mesh), 2 * len(mesh)))

    # Initialize 𝒖 and 𝑭 vectors
    u = np.zeros(K.shape[0])
    F = np.zeros(K.shape[0])
    du = np.zeros(K.shape[0])
    dF = np.zeros(K.shape[0])

    return K, u, F, du, dF


def constraint(mesh, **kwargs):
    r"""
    Assemble the linear constraints.

    The contact model is built using **normal non-penetration conditions** [2]_:

    .. MATH::
        \mathbb{C}_e =
            \left[\begin{array}{rrrr}
                 -1  &  0  &  0  &  0  \\
                  1  &  0  & -1  &  0  \\
                  0  &  0  &  1  &  0  \\
            \end{array}\right]
            \ , \quad \mathbf{H}_e =
            \left(\begin{matrix}
                z_A - \frac{1}{2} \, h_A \\
                z_B - z_A - \frac{1}{2} \, (h_A + h_B) \\
                Z - z_B - \frac{1}{2} \, h_B \\
            \end{matrix}\right)

    where:
        - :math:`z_A` and :math:`z_B` are the vertical positions of nodes A and B.
        - :math:`h_A` and :math:`h_B` are the fiber thicknesses at nodes A and B.

    The vector 𝐟 is the vector of Lagrangian multipliers that corresponds to contact forces.
    It satisfies *KKT conditions*:

    .. MATH::
        \mathbb{C} \, \mathbf{u} \leq \mathbf{H} \, ,
        \quad \mathbf{f} \geq 0
        \quad and \quad \mathbf{f} \, (\mathbf{H} - \mathbb{C} \, \mathbf{u}) = 0

    .. RUBRIC:: Footnotes

    .. [2] `Karush–Kuhn–Tucker conditions, Wikipedia <https://en.wikipedia.org/wiki/Karush%E2%80%93Kuhn%E2%80%93Tucker_conditions>`_.

    Parameters
    ----------
    mesh : pandas.DataFrame
        Fiber mesh represented by a :class:`~.Mesh` object.

    Returns
    -------
    tuple
        C : sparse matrix
            Constraint matrix.
        f : numpy.ndarray
            Force vector.
        H : numpy.ndarray
            Upper-bound vector.
        df : numpy.ndarray
            Incremental force vector.
        dH : numpy.ndarray
            Incremental upper-bound vector.

    Other Parameters
    ----------------
    kwargs :
        Additional keyword arguments ignored by the function.

    """
    # Optional
    if mesh is None:
        mesh = Mesh()

    assert Mesh.check(mesh)

    # Get mesh data
    mask = (mesh.index.values <= mesh.constraint.values)
    i = mesh.index[mask].values
    j = mesh.constraint[mask].values
    k = np.arange(len(i))
    O = i * 0  # : zero
    I = O + 1  # : one

    # Get material data
    mat = mesh.flags.mat
    mesh["h"] = mat.h.loc[mesh.fiber].values
    zi = mesh.z.loc[i].values
    zj = mesh.z.loc[j].values
    hi = mesh.h.loc[i].values
    hj = mesh.h.loc[j].values
    Z = (mesh.z.values + 0.5 * mesh.h.values).max()  # : upper boundary position
    i *= 2
    j *= 2
    k *= 3

    # Create constraint data
    row = np.array([k, k + 1, k + 1, k + 2]).ravel()
    col = np.array([i, i, j, j]).ravel()
    data = np.array([-I, I, -I, I]).ravel()

    # Initialize ℂ matrix
    C = sp.sparse.coo_matrix((data, (row, col)),
                             shape=(3 * len(mesh[mask]), 2 * len(mesh)))

    # Initialize 𝒇 and 𝑯 vectors
    f = np.zeros(C.shape[0])
    H = np.zeros(C.shape[0])
    df = np.zeros(C.shape[0])
    dH = np.zeros(C.shape[0])
    # (X₁ + u₁) ≥ ½h₁ ⟺ -u₁ ≤ X₁ - ½h₁
    H[::3] += zi - 0.5 * hi
    # (X₂ + u₂) - (X₁ + u₁) ≥ ½(h₁ + h₂) ⟺ u₁ - u₂ ≤ X₂ - X₁ - ½(h₁ + h₂)
    H[1::3] += zj - zi - 0.5 * (hi + hj)
    # (X₂ + u₂) ≤ Z - ½h₂ ⟺ u₂ ≤ Z - X₂ - ½h₂
    H[2::3] += Z - zj - 0.5 * hj
    dH[2::3] = 1
    # For end nodes
    H[1::3][mesh[mask].index == mesh[mask].constraint.values] = np.inf

    return C, f, H, df, dH


################################################################################
# Main
################################################################################

if __name__ == "__main__":

    # from fibermat import *

    # Linear model (Ψ² ≫ 1)
    mat = Mat(1, length=1, width=1, thickness=1, shear=1, tensile=np.inf)
    net = Net(mat)
    mesh = Mesh(net)
    print("Linear (Ψ² ≫ 1) =")
    print(4 / np.pi * stiffness(mesh, coupling=1)[0].todense())
    print()

    # Timoshenko model (Ψ² = 1)
    mat = Mat(1, length=1, width=1, thickness=1, shear=2, tensile=2)
    net = Net(mat)
    mesh = Mesh(net)
    print("Timoshenko (Ψ² = 1) = 1 / 2 *")
    print(4 / np.pi * stiffness(mesh, coupling=1)[0].todense())
    print()

    # Euler model (Ψ² ≪ 1)
    mat = Mat(1, length=1, width=1, thickness=1, shear=1e12, tensile=12)
    net = Net(mat)
    mesh = Mesh(net)
    print("Euler (Ψ² ≪ 1) = 1 / 12 *")
    print(4 / np.pi * stiffness(mesh, coupling=1)[0].todense())
    print()

    # Generate a set of fibers
    mat = Mat(100)
    # Build the fiber network
    net = Net(mat)
    # Stack fibers
    stack = Stack(net)
    # Create the fiber mesh
    mesh = Mesh(stack)

    # Assemble the quadratic programming system
    K, u, F, du, dF = stiffness(mesh)
    C, f, H, df, dH = constraint(mesh)
    P = sp.sparse.bmat([[K, C.T], [C, None]], format='csc')
    # Permutation of indices
    perm = sp.sparse.csgraph.reverse_cuthill_mckee(P, symmetric_mode=True)
    # Visualize the system
    fig, ax = plt.subplots(1, 2, figsize=(2 * 6.4, 4.8))
    plot_system((K, u, F, du, dF), (C, f, H, df, dH), perm=None, ax=ax[0])
    plot_system((K, u, F, du, dF), (C, f, H, df, dH), perm=perm, ax=ax[1])
    plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import warnings
from matplotlib import pyplot as plt
from tqdm import tqdm

from fibermat import *
from fibermat import Mat, Net


class Mesh(pd.DataFrame):
    """
    A class inherited from pandas.DataFrame_ to **represent a mesh structure** for a set of discontinuous fibers.

    It defines:

        - the *beam* elements (intra-fiber connections).
        - the *constraint* elements (inter-fiber connections).

    Parameters
    ----------
    net : pandas.DataFrame, optional
         Fiber network represented by a :class:`~.Net` object.

    Other Parameters
    ----------------
    kwargs :
        Additional keyword arguments ignored by the function.

    .. NOTE::
        The constructor calls :meth:`init` method if the object is instantiated with parameters.
        Otherwise, initialization is performed with the pandas.DataFrame_ constructor.

    .. _pandas.DataFrame: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html

    :Use:

        >>> # Generate a set of fibers
        >>> mat = Mat(100)
        >>> # Build the fiber network
        >>> net = Net(mat)
        >>> # Create the fiber mesh
        >>> mesh = Mesh(net)
        >>> mesh
              fiber          s          x          y          z  beam  constraint
        0         0 -12.500000  -1.176401  -3.074404 -24.338157     1           0
        1         0 -11.222534  -0.806746  -1.851590 -24.338157     2         752
        2         0 -10.466114  -0.587864  -1.127531 -24.338157     3         123
        3         0 -10.009779  -0.455816  -0.690719 -24.338157     4        1519
        4         0  -5.432013   0.868835   3.691203 -24.338157     5         706
        ...     ...        ...        ...        ...        ...   ...         ...
        1729     99   6.159453 -21.379789  -8.424815  24.516947  1730        1094
        1730     99   6.740970 -21.060169  -8.910618  24.516947  1731         157
        1731     99   7.284437 -20.761462  -9.364634  24.516947  1732        1294
        1732     99  11.270660 -18.570503 -12.694751  24.516947  1733        1585
        1733     99  12.500000 -17.894817 -13.721749  24.516947  1716        1733
        <BLANKLINE>
        [1734 rows x 7 columns]

    Data
    ----
    + index : pandas.Index
        Node label. Each label refers to a unique node.
    + Node:
        - fiber : pandas.Series
            Label of the fiber to which the node belongs
        - s : pandas.Series
            Node curvilinear abscissa along the fiber (mm).
    + Node position:
        - x : pandas.Series
            X-coordinate of the node (mm).
        - y : pandas.Series
            Y-coordinate of the node (mm).
        - z : pandas.Series
            Z-coordinate of the node (mm).
    + Elements:
        - constraint : pandas.Series
            Index of the connected node. It defines a constraint on the relative node positions.
        - beam : pandas.Series
            Index of the next node along the fiber. It defines a mechanical beam element.

    ----

    Attributes
    ----------
    :attr:`attrs` :
        Global attributes of DataFrame.

    Methods
    -------
    :meth:`init` :
        Create a discontinuous fiber mesh.
    :meth:`check` :
        Check that a :class:`Mesh` object is defined correctly.

    ----

    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the `Mesh` object.

        Parameters
        ----------
        args :
            Additional positional arguments passed to the constructor.
        kwargs :
            Additional keyword arguments passed to the constructor.

        See also
        --------
        `Mesh.init`.

        """
        if (len(args) and isinstance(args[0], pd.DataFrame)
                and not isinstance(args[0], Net)):
            # Initialize the DataFrame from argument
            super().__init__(*args, **kwargs)
            # Copy global attributes from argument
            self.attrs = args[0].attrs
            # Copy global flags from argument
            self.flags.mat = args[0].flags.mat

        else:
            # Initialize the DataFrame from parameters
            self.__init__(Mesh.init(*args, **kwargs))

        # Check `Mesh` object
        assert Mesh.check(self)

    # ~~~ Constructor ~~~ #

    @staticmethod
    def init(net=None, **kwargs):
        """
        Create a discontinuous fiber mesh.

        Parameters
        ----------
        net : pandas.DataFrame, optional
             Fiber network represented by a :class:`~.Net` object.

        Returns
        -------
        mesh : pandas.DataFrame
            Initialized :class:`Mesh` object.

        Other Parameters
        ----------------
        kwargs :
            Additional keyword arguments ignored by the function.

        """
        # Optional
        if net is None:
            net = Net()

        assert Net.check(net)

        # Get network data
        pairs = net[[*"AB"]].values.ravel()
        abscissa = net[["sA", "sB"]].values.ravel()
        points = net[["xA", "yA", "zA", "xB", "yB", "zB"]].values.reshape(-1, 3)

        # Initialize mesh DataFrame
        mesh = pd.DataFrame(
            data=np.c_[pairs, abscissa, points],
            columns=["fiber", *"sxyz"]
        )
        mesh["beam"] = 0
        mesh["constraint"] = 0
        # Convert type to int
        mesh.fiber = mesh.fiber.astype(int)

        # Set attributes
        mesh.attrs = net.attrs

        # Set flags
        mesh.flags.mat = net.flags.mat

        if len(net):
            # Inter-fiber connections
            mesh.constraint = mesh.index.values.reshape(-1, 2)[:, ::-1].ravel()
            # Sort nodes along fibers
            mesh.sort_values(by=["fiber", "s"], inplace=True)
            # Intra-fiber elements
            mesh.beam = np.hstack(
                mesh.groupby("fiber")
                .apply(lambda x: np.roll(x.index, -1))
            )
            # Reorder nodes
            indices = np.argsort(mesh.index).astype(int)
            mesh.index = indices[mesh.index]
            mesh.beam = indices[mesh.beam]
            mesh.constraint = indices[mesh.constraint]
            # Correct end nodes
            mask = (mesh.fiber.values == mesh.fiber.loc[mesh.constraint].values)
            mesh.constraint.values[mask] = mesh.index[mask]

        # Return the `Mesh` object
        return mesh

    # ~~~ Public properties ~~~ #

    @property
    def attrs(self):
        """
        Global attributes of DataFrame:
            - n : int
                Number of fibers. By default, it is empty (n = 0).
            - size : float
                Box dimensions (mm). By default, the domain is a 50 mm square cube.
            - periodic : bool
                Boundary periodicity. By default, the domain is periodic.
            - threshold : float, optional
                Threshold distance value for proximity detection (mm).

        """
        return self._attrs

    @attrs.setter
    def attrs(self, attrs):
        self._attrs = attrs

    # ~~~ Public methods ~~~ #

    def check(self):
        """
        Check that a :class:`Mesh` object is defined correctly.

        This method is called when a :class:`Mesh` object is initialized.

        Raises
        ------
        KeyError
            If any keys are missing from the columns of :class:`Mesh` object.
        AttributeError
            If any attributes are missing from the dictionary :attr:`attrs`.
        IndexError
            If row indices are incorrectly defined:
                - Row indices are not unique in [0,..., m-1] where m is the number of nodes.
                - Node labels are not sorted.
        TypeError
            If labels are not integers.
        ValueError
            If fibers, beams or constraints are ill-defined, improperly connected or not sorted.

        Returns
        -------
        bool
            Indicates whether the object can be instantiated as :class:`Mesh`.

        .. TIP::
            - If `self` is None, it returns an empty :class:`Mesh` object.
            - If a `"skip_check"` flag is True in :attr:`attrs`, the check is passed.

        """
        if self is None:
            mesh = Mesh()
        else:
            mesh = self

        if "skip_check" in mesh.attrs.keys() and mesh.attrs["skip_check"]:
            warnings.warn("{}.attrs['skip_check'] is active."
                          " Delete it or set it to False.".format(mesh.__class__),
                          UserWarning)
            return True

        assert Mat.check(mesh.flags.mat)

        # Keys
        try:
            mesh[["fiber", *"sxyz", "beam", "constraint"]]
        except KeyError as e:
            raise KeyError(e)

        # Attributes
        if not ("n" in mesh.attrs.keys()):
            raise AttributeError("'n' is not in attribute dictionary.")
        if not ("size" in mesh.attrs.keys()):
            raise AttributeError("'size' is not in attribute dictionary.")
        if not ("periodic" in mesh.attrs.keys()):
            raise AttributeError("'periodic' is not in attribute dictionary.")

        # Indices
        if not np.all(np.unique(mesh.index) == np.arange(len(mesh))):
            raise IndexError("Row indices must be unique in [0,..., {}]."
                             .format(len(mesh) - 1))
        if not np.all(mesh.index == np.arange(len(mesh))):
            raise IndexError("Node labels must be sorted.")

        # Types
        if mesh["fiber"].values.dtype != int:
            raise TypeError("Fiber labels are not integers.")
        if mesh[["beam", "constraint"]].values.dtype != int:
            raise TypeError("Node labels are not integers.")

        # Data
        if len(mesh):
            if not (0 <= mesh.fiber.min()
                    and mesh.fiber.max() < mesh.attrs["n"]):
                raise ValueError("Fiber labels must be in [0,..., {}]."
                                 .format(mesh.attrs["n"] - 1))
            if not (0 <= mesh.beam.min() and mesh.beam.max() < len(mesh)):
                raise ValueError("Beam labels must be in [0,..., {}]."
                                 .format(len(mesh) - 1))
            if not (0 <= mesh.constraint.min()
                    and mesh.constraint.max() < len(mesh)):
                raise ValueError("Constraint labels must be in [0,..., {}]."
                                 .format(len(mesh) - 1))
            if not np.all(np.sort(mesh.fiber) == mesh.fiber):
                raise ValueError("Fibers are not sorted.")
            if not np.all(mesh.sort_values(by=["fiber", "s"]).index
                          == mesh.index):
                raise ValueError("Nodes are not ordered along fibers.")
            if not np.all(mesh.beam == np.hstack(
                    mesh.groupby("fiber").apply(lambda x: np.roll(x.index, -1))
            )):
                raise ValueError("Beams are not sorted.")
            if not np.all(mesh.constraint[mesh.constraint].values == mesh.index):
                raise ValueError("Constraints are not reciprocal.")
            end1 = mesh.loc[(mesh.groupby("fiber")
                             .apply(lambda x: x.index[0]))]
            end2 = mesh.loc[(mesh.groupby("fiber")
                             .apply(lambda x: x.index[-1]))]
            if (not np.all(end1.constraint == end1.index)
                    or not np.all(end2.constraint == end2.index)):
                raise ValueError(
                    "End node constraints must be equal to node labels.")

        # Return True if the test is correct
        return True

    def _is(self):
        if self is None:
            return False
        try:
            __class__.check(self)
            return True
        except:
            return False


################################################################################
# Main
################################################################################

if __name__ == "__main__":

    # from fibermat import *

    # Generate a set of fibers
    mat = Mat(10)
    # Build the fiber network
    net = Net(mat)
    # Create the fiber mesh
    mesh = Mesh(net)

    # Check data
    Mesh.check(mesh)  # or `mesh.check()`
    # -> returns True if correct, otherwise it raises an error.

    # Figure
    fig, ax = plt.subplots(subplot_kw=dict(projection='3d', aspect='equal',
                                           xlabel="X", ylabel="Y", zlabel="Z"))
    ax.view_init(azim=45, elev=30, roll=0)
    if len(mesh):
        # Draw elements
        for i, j, k in tqdm(zip(mesh.index, mesh.beam, mesh.constraint),
                            total=len(mesh), desc="Draw mesh"):
            # Get element data
            a, b, c = mesh.iloc[[i, j, k]][[*"xyz"]].values
            if mesh.iloc[i].s < mesh.iloc[j].s:
                # Draw intra-fiber connection
                plt.plot(*np.c_[a, b],
                         c=plt.cm.tab10(mesh.fiber.iloc[i] % 10))
            if mesh.iloc[i].z < mesh.iloc[k].z:
                # Draw inter-fiber connection
                plt.plot(*np.c_[a, c], '--ok',
                         lw=1, mfc='none', ms=3, alpha=0.2)
            if mesh.iloc[i].fiber == mesh.iloc[k].fiber:
                # Draw fiber end nodes
                plt.plot(*np.c_[a, c], '+k', ms=3, alpha=0.2)
    # Set drawing box dimensions
    ax.set_xlim(-0.5 * mesh.attrs["size"], 0.5 * mesh.attrs["size"])
    ax.set_ylim(-0.5 * mesh.attrs["size"], 0.5 * mesh.attrs["size"])
    plt.show()



import numpy as np
import pandas as pd

from fibermat import *


################################################################################
# Tests
################################################################################

def test_solver():
    """
    Test the solver for a problem of 3 fibers.

    """
    mat = Mat(3)
    mat.x = -5, 0, 5
    mat.y = 0, 0, 0
    mat.z = 0.5, 1.5, 2.5
    mat.u = 0, 1, 0
    mat.v = 1, 0, 1
    mat.G = 4 / np.pi * 10

    net = Net(mat)
    mesh = Mesh(net)

    # Solve the mechanical packing problem
    K, C, u, f, F, H, Z, rlambda, mask, err = solve(
        mesh,
        stiffness(mesh),
        constraint(mesh),
        packing=4,
    )

    assert np.allclose(
        f(1) @ C,
        np.array([
             0., 0.,  # Fiber 0
             0., 0.,  # Fiber 0
             0., 0.,  # Fiber 0
             0., 0.,  # Fiber 1
            -1., 0.,  # Fiber 1 (upper)
             1., 0.,  # Fiber 1 (lower)
             0., 0.,  # Fiber 1
             0., 0.,  # Fiber 2
             0., 0.,  # Fiber 2
             0., 0.,  # Fiber 2
        ])
    )


################################################################################
# Main
################################################################################

if __name__ == '__main__':

    test_solver()

    mat = Mat(3)
    mat.x = -5, 0, 5
    mat.y = 0, 0, 0
    mat.z = 0.5, 1.5, 2.5
    mat.u = 0, 1, 0
    mat.v = 1, 0, 1
    mat.G = 4 / np.pi * 10

    net = Net(mat)
    mesh = Mesh(net)

    # Solve the mechanical packing problem
    K, C, u, f, F, H, Z, rlambda, mask, err = solve(
        mesh,
        stiffness(mesh),
        constraint(mesh),
        packing=4,
    )

    # Deform the mesh
    mesh.z += u(1)[::2]

    # Figure
    fig, ax = plt.subplots(subplot_kw=dict(projection='3d', aspect='equal',
                                           xlabel="X", ylabel="Y", zlabel="Z"))
    ax.view_init(azim=45, elev=30, roll=0)
    if len(mesh):
        # Draw elements
        for i, j, k in tqdm(zip(mesh.index, mesh.beam, mesh.constraint),
                            total=len(mesh), desc="Draw mesh"):
            # Get element data
            a, b, c = mesh.iloc[[i, j, k]][[*"xyz"]].values
            if mesh.iloc[i].s < mesh.iloc[j].s:
                # Draw intra-fiber connection
                plt.plot(*np.c_[a, b],
                         c=plt.cm.tab10(mesh.fiber.iloc[i] % 10))
            if mesh.iloc[i].z < mesh.iloc[k].z:
                # Draw inter-fiber connection
                plt.plot(*np.c_[a, c], '--ok',
                         lw=1, mfc='none', ms=3, alpha=0.2)
            if mesh.iloc[i].fiber == mesh.iloc[k].fiber:
                # Draw fiber end nodes
                plt.plot(*np.c_[a, c], '+k', ms=3, alpha=0.2)
    # Set drawing box dimensions
    ax.set_xlim(-0.5 * mesh.attrs["size"], 0.5 * mesh.attrs["size"])
    ax.set_ylim(-0.5 * mesh.attrs["size"], 0.5 * mesh.attrs["size"])
    plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
🗹 Widgets
----------

Classes
-------
PipButton :
    Display a button widget to install Python packages in Jupyter notebooks.
MatplotlibButton :
    Display a button widget to toggle between matplotlib inline and qt display modes in Jupyter notebooks.

"""

import IPython
import ipywidgets
import os
import subprocess
from IPython.display import display, HTML


class PipButton(ipywidgets.Button):
    """
    A class to install Python packages in Jupyter notebooks.

    Parameters
    ----------
    packages : str,...
        Names of Python packages to install.

    Examples
    --------
    >>> pipButton = PipButton("matplotlib", "numpy", "pandas")
    >>> pipButton()

    Methods
    -------
    PipButton(*packages)
        Initialize a `PipButton` button widget.
    __call__()
        Display widget.

    """

    def __init__(self, *packages, option=None, **kwargs):
        """
        Initialize a `PipButton` button widget.

        Parameters
        ----------
        packages : str,...
            Names of Python packages to install.

        Other Parameters
        ----------------
        option : str, optional
            Additional options passed to `pip install` command.
        kwargs :
            Additional Python packages to install. The keys are passed to `pip install` and the values are used for `import`.

        """
        # Output stream
        output = ipywidgets.Output()

        pip_packages = [*packages, *kwargs.keys()]
        py_packages = [*packages, *kwargs.values()]

        if option is None:
            option = ""

        @output.capture(clear_output=True)
        def is_clicked(*args):
            self.button_style = 'primary'
            self.description = "Running..."
            # Installation command line
            result = os.system("pip install -q {} {}"
                               .format(option, " ".join(pip_packages)))
            if result == 0:
                # -> If successful
                self.button_style = 'success'
                self.description = "Installed"
                # Get package version
                cmd = 'python -Wignore -c "{}"'.format(
                    "".join([
                        "from {0} import __version__; print(__version__);"
                        .format(package) for package in py_packages
                    ])
                )
                versions = (subprocess.check_output(cmd, shell=True)
                            .decode()
                            .split('\n'))
                # Display package installation
                display(HTML(
                    "<table><thead><tr><th>Package</th><th>Version</th>"
                    + "</tr></thead><tbody>"
                    + "".join([
                        "<tr>"
                        "<th>{0}</th>"
                        "<th><a href='https://pypi.org/project/{0}'>{1}</a>"
                        "</tr>"
                        .format(package, version)
                        for package, version in zip(pip_packages, versions)])
                    + "</tbody></table>"
                ))
            else:
                # -> If unsuccessful
                self.button_style = 'danger'
                self.description = "Not Installed"

        # Initialize `PipButton` widget
        super().__init__()
        self.description = "Install packages"
        # Click events
        self.on_click(is_clicked)
        self.output = output

    # ~~~ Private methods ~~~ #

    def __call__(self):
        """Display widget."""
        display(self)
        display(self.output)


class MatplotlibButton(ipywidgets.Button):
    """
    A class to display a widget for switching between:
        - `%matplotlib inline`
        - `%matplotlib qt`
    in Jupyter notebooks.

    Examples
    --------
    >>> matplotlibButton = MatplotlibButton()
    >>> if "matplotlibButton" not in globals():
    >>>     matplotlibButton()

    Methods
    -------
    MatplotlibButton()
        Initialize a `MatplotlibButton` button widget.
    __call__()
        Display widget.

    """

    def __init__(self, **kwargs):
        """
        Initialize a `MatplotlibButton` widget.

        Parameters
        ----------
        kwargs :
            Additional keyword arguments to pass to `Button` constructor.

        """
        def is_clicked(*args):
            if self.description == "%matplotlib inline":
                self.description = "%matplotlib qt"
                self.button_style = 'success'
                IPython.get_ipython().run_cell("%matplotlib qt")
            else:
                self.description = "%matplotlib inline"
                self.button_style = 'primary'
                IPython.get_ipython().run_cell("%matplotlib inline")
            self.value = self.description

        # Initialize `MatplotlibButton` widget
        super().__init__(**kwargs)
        is_clicked()
        # Click events
        self.on_click(is_clicked)

    # ~~~ Private methods ~~~ #

    def __call__(self):
        """Display widget."""
        display(self)


################################################################################
# Main
################################################################################

if __name__ == "__main__":

    pipButton = PipButton(
        "fibermat",
        "matplotlib",
        "numpy",
        "pandas",
        "pyvista",
        "scipy",
        **{"scikit-learn": "sklearn"},
        tqdm="tqdm",
        option="--upgrade"
    )
    pipButton()

    if "matplotlibButton" not in globals():
        matplotlibButton = MatplotlibButton()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from matplotlib import pyplot as plt
from scipy.interpolate import interp1d

from fibermat import *


class Interpolate(interp1d):
    """
    A class for interpolating array values inherited from `scipy.interpolate.interp1d
    <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html>`_.

    Parameters
    ----------
    y : array-like
        Data to be interpolated.

    Other Parameters
    ----------------
    size : int, optional
        Number of points used for interpolation.
    kwargs :
        Additional keyword arguments passed to `interp1d` constructor.

    Attributes
    ----------
    t : array-like
        Interpolation parameter between 0 and 1.
    dtype : type
        Data type used for interpolation.

    Methods
    -------
    :meth:`__call__` :
        Return interpolated data.

    ----

    """

    def __init__(self, y, size=None, **kwargs):
        """
        Initialize an `Interpolation` object.

        Parameters
        ----------
        y : array-like
            Data to be interpolated.

        Other Parameters
        ----------------
        size : int, optional
            Number of points used for interpolation.
        kwargs :
            Additional keyword arguments passed to `interp1d` constructor.

        """
        t = np.linspace(0, 1, len(y))
        f = interp1d(t, y, axis=0,
                     bounds_error=False, fill_value=y[-1], **kwargs)
        if size is not None:
            t = np.linspace(0, 1, size)
        # Initialize interpolation
        super().__init__(t, f(t), axis=0,
                         bounds_error=False, fill_value=y[-1], **kwargs)
        # Set attributes
        self.t = t
        self.dtype = np.array(y).dtype

    def __call__(self, t=None):
        """
        Return interpolated data.

        Parameters
        ----------
        t : array-like, optional
            Interpolation parameter between 0 and 1.

        Returns
        -------
        array-like
            Interpolated data.

        """
        if t is None:
            t = self.t
        return super().__call__(t).astype(self.dtype)


################################################################################
# Main
################################################################################

if __name__ == "__main__":

    # from fibermat import *

    # Reference solution
    x = np.linspace(0, 10, 1001)
    y = np.sin(x)

    # Interpolation nodes
    X = x[::100]
    Y = y[::100]

    # Interpolated functions
    f_ = Interpolate(Y, size=11, kind='previous')
    g_ = Interpolate(Y, size=11, kind='linear')
    h_ = Interpolate(Y, size=11, kind='cubic')
    x_ = Interpolate(X, size=11)
    t = np.linspace(0, 1, 100)

    # Figure
    plt.figure()
    p, = plt.plot(x, y, label="Reference solution")
    plt.plot(X, Y, 'o', color=p.get_color(), zorder=np.inf, label="Interpolation nodes")
    p, = plt.plot(x_(t), f_(t), '--')
    plt.plot(x_(), f_(), 'o', ms=10, mfc='none', c=p.get_color())
    plt.plot([], [], 'o--', mfc='none', c=p.get_color(), label="Previous")
    p, = plt.plot(x_(t), g_(t), '--')
    plt.plot(x_(), g_(), 'o', ms=14, mfc='none', c=p.get_color())
    plt.plot([], [], 'o--', mfc='none', c=p.get_color(), label="Linear")
    p, = plt.plot(x_(t), h_(t), '--')
    plt.plot(x_(), h_(), 'o', ms=18, mfc='none', c=p.get_color())
    plt.plot([], [], 'o--', mfc='none', c=p.get_color(), label="Cubic")
    plt.xlabel("X")
    plt.ylabel("Y")
    plt.legend()
    plt.title("Interpolations")
    plt.show()



from fibermat import *

mesh = Mesh(Stack(Net(Mat(100))))

K, C, u, f, F, H, Z, rlambda, mask, err = solve(
    mesh,
    stiffness(mesh),
    constraint(mesh),
    packing=4
)

vtk_mesh(mesh, displacement(u(1))).plot()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import scipy as sp
import warnings
from matplotlib import pyplot as plt
from tqdm import tqdm

from fibermat import *
from fibermat.model.timoshenko import stiffness, constraint
from fibermat.utils.interpolation import Interpolate


def solve(mesh, stiffness, constraint, packing=1., itermax=1000,
          solve=sp.sparse.linalg.spsolve, perm=None, tol=1e-6,
          errtol=1e-6, interp_size=None, verbose=True, **kwargs):
    r"""
    An iterative mechanical solver for **fiber packing problems**.

    It solves the *quadratic programming problem*:

    .. MATH::
        \min_{\mathbf{u}, \mathbf{f}} \left(
            \frac{1}{2} \, \mathbf{u} \, \mathbb{K} \, \mathbf{u}
            - \mathbf{F} \, \mathbf{u}
            - \mathbf{f} \, (\mathbf{H} - \mathbb{C} \, \mathbf{u})
        \right)
    .. MATH::
        \quad s.t. \quad \mathbb{C} \, \mathbf{u} \leq \mathbf{H} \, ,
        \quad \mathbf{u} \leq 0 \, ,
        \quad \mathbf{f} \geq 0
        \quad and
        \quad \mathbf{f} \, (\mathbf{H} - \mathbb{C} \, \mathbf{u}) = 0

    where:
        - 𝐮 is the vector of generalized displacements (*unknowns of the problem*).
        - 𝐟 is the vector of generalized forces (*unknown Lagrange multipliers*).
        - 𝕂 is the stiffness matrix of the fiber set.
        - 𝐅 is the vector of external efforts.
        - ℂ is the matrix of non-penetration constraints.
        - 𝐇 is the vector of minimal distances between fibers (minimal distances).

    The *mechanical equilibrium* allows reformulating the problem as a system of inequalities:

    .. MATH::
        \Rightarrow \quad \left[\begin{matrix}
            \mathbb{K} & \mathbb{C}^T \\
            \mathbb{C} & 0
        \end{matrix}\right] \binom{\mathbf{u}}{\mathbf{f}}
            \leq \binom{\mathbf{F}}{\mathbf{H}}

    which is solved using an iterative *Updated Lagrangian Approach*.

    .. HINT::
        Models used to build the matrices are implemented in :ref:`🔧 Model`:
            - 𝕂 and 𝐅 : :func:`~.model.timoshenko.stiffness`.
            - ℂ and 𝐇 : :func:`~.model.timoshenko.constraint`.

    Parameters
    ----------
    mesh : pandas.DataFrame, optional
        Fiber mesh represented by a :class:`~.Mesh` object.
    stiffness : tuple
        K : sparse matrix
            Stiffness matrix (symmetric positive-semi definite).
        u : numpy.ndarray
            Displacement vector.
        F : numpy.ndarray
            Load vector.
        du : numpy.ndarray
            Incremental displacement vector.
        dF : numpy.ndarray
            Incremental load vector.
    constraint : tuple
        C : sparse matrix
            Constraint matrix.
        f : numpy.ndarray
            Force vector.
        H : numpy.ndarray
            Upper-bound vector.
        df : numpy.ndarray
            Incremental force vector.
        dH : numpy.ndarray
            Incremental upper-bound vector.
    packing : float, optional
        Targeted value of packing. Must be greater than 1. Default is 1.0.
    itermax : int, optional
        Maximum number of solver iterations. Default is 1000.

    Returns
    -------
    tuple
        K : sparse matrix
            Stiffness matrix (symmetric positive-semi definite).
        C : sparse matrix
            Constraint matrix.
        u : Interpolate
            Displacement vector.
        f : Interpolate
            Force vector.
        F : Interpolate
            Load vector.
        H : Interpolate
            Upper-bound vector.
        Z : Interpolate
            Upper-boundary position.
        rlambda : Interpolate
            Compaction stretch factor.
        mask : Interpolate
            Active rows and columns in the system of inequations.
        err : Interpolate
            Numerical error of the linear solver.

    .. SEEALSO::
        Simulation results are given as functions of a pseudo-time parameter (between 0 and 1) using :class:`~.interpolation.Interpolate` objects.

    Other Parameters
    ----------------
    solve : callable, optional
        Sparse solver. It is a callable object that takes as inputs a sparse
        symmetric matrix 𝔸 and a vector 𝐛 and returns the solution 𝐱 of the
        linear system: 𝔸 𝐱 = 𝐛. Default is `scipy.sparse.linalg.spsolve`.
    perm : numpy.ndarray, optional
        Permutation of indices.
    tol : float, optional
        Tolerance used for contact detection (mm). Default is 1e-6 mm.
    errtol : float, optional
        Tolerance for the numerical error. Default is 1e-6.
    interp_size : int, optional
        Size of array used for interpolation. Default is None.
    verbose : bool, optional
        If True, a progress bar is displayed. Default is True.
    kwargs :
        Additional keyword arguments ignored by the function.

    :Use:

        >>> # Generate a set of fibers
        >>> mat = Mat(100)
        >>> # Build the fiber network
        >>> net = Net(mat)
        >>> # Create the fiber mesh
        >>> mesh = Mesh(net)
        >>> # Solve the mechanical packing problem
        >>> K, C, u, f, F, H, Z, rlambda, mask, err = solve(mesh, stiffness(mesh), constraint(mesh), packing=4)

    """
    # Optional
    if mesh is None:
        mesh = Mesh()

    # Assemble the quadratic programming system
    K, u, F, du, dF = stiffness
    C, f, H, df, dH = constraint
    P = sp.sparse.bmat([[K, C.T], [C, None]], format='csc')
    x = np.r_[u, f]
    q = np.r_[F, H]
    dx = np.r_[du, df]
    dq = np.r_[dF, dH]

    if perm is None:
        perm = np.arange(P.shape[0])

    u, f = np.split(x, [K.shape[0]])  # Memory-shared
    F, H = np.split(q, [K.shape[0]])  # Memory-shared
    du, df = np.split(dx, [K.shape[0]])  # Memory-shared
    dF, dH = np.split(dq, [K.shape[0]])  # Memory-shared

    u_ = [u.copy()]
    f_ = [f.copy()]
    F_ = [F.copy()]
    H_ = [H.copy()]
    Z_ = [(mesh.z.values + 0.5 * mesh.h.values).max()]
    rlambda_ = [1.0]
    mask_ = [(np.real(q - P @ x) <= tol)]
    err_ = [0]

    # Incremental solver
    with tqdm(total=itermax, desc="Packing: {:.2}".format(rlambda_[-1]),
              disable=not verbose) as pbar:
        i = 0
        while (i < pbar.total) and (rlambda_[-1] < packing):
            # Solve step
            dx *= 0
            # mask = [True] * K.shape[0] + [*(C @ u >= H - tol)]
            mask = (np.real(q - P @ x) <= tol)
            mask &= np.array(np.sum(np.abs(np.real(P)), axis=0) > 0).ravel()
            # Solve linear problem
            # dx[mask] += np.real(solve(P[np.ix_(mask, mask)], dq[mask]))
            dx[perm[mask[perm]]] = np.real(
                solve(
                    P[np.ix_(perm[mask[perm]], perm[mask[perm]])],
                    dq[perm[mask[perm]]]
                )
            )
            # Calculate error
            err = np.linalg.norm(P[np.ix_(mask, mask)] @ dx[mask] - dq[mask])
            # Calculate evolution
            d = np.real(q - P @ x)
            v = np.real(dq - P @ dx)

            try:
                # Calculate the next step
                dU = -min(d[(d > tol) & (v > 0)] / v[(d > tol) & (v > 0)])
                # Stopping criteria
                stop = False
                if err > errtol:
                    if verbose:
                        warnings.warn("Stopping criteria: err = {}".format(err),
                                      UserWarning)
                    stop = True
                if stop:
                    raise ValueError
            except ValueError:
                break

            # Jump to the next step
            x += dx * dU
            q += dq * dU
            # Store results
            u_ += [u.copy()]
            f_ += [f.copy()]
            F_ += [F.copy()]
            H_ += [H.copy()]
            Z_ += [Z_[-1] + dU]
            rlambda_ += [Z_[0] / Z_[-1]]
            mask_ += [mask.copy()]
            err_ += [err.copy()]

            # Update
            i += 1
            pbar.set_description("Packing: {:.2}".format(rlambda_[-1]))
            pbar.update()

    # Interpolate results
    with warnings.catch_warnings():
        # Ignore warning messages due to infinite values in 𝐇
        warnings.filterwarnings('ignore')
        u = Interpolate(u_, size=interp_size)
        f = Interpolate(f_, size=interp_size)
        F = Interpolate(F_, size=interp_size)
        H = Interpolate(H_, size=interp_size)
        Z = Interpolate(Z_, size=interp_size)
        rlambda = Interpolate(rlambda_)
        mask = Interpolate(mask_, kind='previous')
        err = Interpolate(err_, kind='previous')

    # Return interpolated results
    return K, C, u, f, F, H, Z, rlambda, mask, err


def plot_system(stiffness, constraint,
                solve=sp.sparse.linalg.spsolve, perm=None, tol=1e-6, ax=None):
    """
    Visualize the system of equations and calculate the step error.

    Parameters
    ----------
    stiffness : tuple
        K : sparse matrix
            Stiffness matrix (symmetric positive-semi definite).
        u : numpy.ndarray
            Displacement vector.
        F : numpy.ndarray
            Load vector.
        du : numpy.ndarray
            Incremental displacement vector.
        dF : numpy.ndarray
            Incremental load vector.
    constraint : tuple
        C : sparse matrix
            Constraint matrix.
        f : numpy.ndarray
            Force vector.
        H : numpy.ndarray
            Upper-bound vector.
        df : numpy.ndarray
            Incremental force vector.
        dH : numpy.ndarray
            Incremental upper-bound vector.

    Returns
    -------
    ax : matplotlib.axes.Axes
        Matplotlib axes.

    Other Parameters
    ----------------
    solve : callable, optional
        Sparse solver. It is a callable object that takes as inputs a sparse
        symmetric matrix 𝔸 and a vector 𝒃 and returns the solution 𝒙 of the
        linear system: 𝔸 𝒙 = 𝒃. Default is `scipy.sparse.linalg.spsolve`.
    perm : numpy.ndarray, optional
        Permutation of indices.
    tol : float, optional
        Tolerance used for contact detection (mm). Default is 1e-6 mm.
    ax : matplotlib.axes.Axes, optional
        Matplotlib axes.

    """
    # Assemble the quadratic programming system
    K, u, F, du, dF = stiffness
    C, f, H, df, dH = constraint
    P = sp.sparse.bmat([[K, C.T], [C, None]], format='csc')
    x = np.r_[u, f]
    q = np.r_[F, H]
    dx = np.r_[du, df]
    dq = np.r_[dF, dH]

    mask0 = np.array([True] * K.shape[0] + [False] * C.shape[0])
    D0 = sp.sparse.diags(mask0.astype(float))
    mask1 = np.array([False] * K.shape[0] + [*(np.isclose(C @ u, H) & np.tile([True, False, False], C.shape[0] // 3))])
    D1 = sp.sparse.diags(mask1.astype(float))
    mask2 = np.array([False] * K.shape[0] + [*(np.isclose(C @ u, H) & np.tile([False, True, False], C.shape[0] // 3))])
    D2 = sp.sparse.diags(mask2.astype(float))
    mask3 = np.array([False] * K.shape[0] + [*(np.isclose(C @ u, H) & np.tile([False, False, True], C.shape[0] // 3))])
    D3 = sp.sparse.diags(mask3.astype(float))
    mask4 = np.array([False] * K.shape[0] + [*(~np.isclose(C @ u, H))])
    D4 = sp.sparse.diags(mask4.astype(float))

    if perm is None:
        perm = np.arange(P.shape[0])

    # Figure
    if ax is None:
        fig, ax = plt.subplots()
    ax.spy((D0 @ P @ D0)[np.ix_(perm, perm)], ms=3, color='black', alpha=0.5, label="stiffness")
    ax.spy((D2 @ P + P @ D2)[np.ix_(perm, perm)], ms=3, color='tab:blue', alpha=0.25, label="inner")
    ax.spy((D1 @ P + P @ D1)[np.ix_(perm, perm)], ms=3, color='tab:green', alpha=0.5, label="lower")
    ax.spy((D3 @ P + P @ D3)[np.ix_(perm, perm)], ms=3, color='tab:red', alpha=0.5, label="upper")
    ax.spy((D4 @ P + P @ D4)[np.ix_(perm, perm)], ms=1, color='gray', zorder=-1, alpha=0.1,
           label="inactive")
    ax.legend()

    mask = (np.real(q - P @ x) <= tol)
    mask &= np.array(np.sum(np.abs(np.real(P)), axis=0) > 0).ravel()

    # Solve linear problem
    dx[perm[mask[perm]]] = np.real(
        solve(
            P[np.ix_(perm[mask[perm]], perm[mask[perm]])],
            dq[perm[mask[perm]]]
        )
    )

    # Calculate error
    err = np.linalg.norm(P[np.ix_(mask, mask)] @ dx[mask] - dq[mask])
    print(err)

    return ax


################################################################################
# Main
################################################################################

if __name__ == "__main__":

    # from fibermat import *

    # Generate a set of fibers
    mat = Mat(10)
    # Build the fiber network
    net = Net(mat)
    # Stack fibers
    stack = Stack(net)
    # Create the fiber mesh
    mesh = Mesh(stack)

    # Solve the mechanical packing problem
    K, C, u, f, F, H, Z, rlambda, mask, err = solve(
        mesh,
        stiffness(mesh),
        constraint(mesh),
        packing=4,
    )

    # Deform the mesh
    mesh.z += displacement(u(1))

    # Figure
    fig, ax = plt.subplots(subplot_kw=dict(projection='3d', aspect='equal',
                                           xlabel="X", ylabel="Y", zlabel="Z"))
    ax.view_init(azim=45, elev=30, roll=0)
    if len(mesh):
        # Draw elements
        for i, j, k in tqdm(zip(mesh.index, mesh.beam, mesh.constraint),
                            total=len(mesh), desc="Draw mesh"):
            # Get element data
            a, b, c = mesh.iloc[[i, j, k]][[*"xyz"]].values
            if mesh.iloc[i].s < mesh.iloc[j].s:
                # Draw intra-fiber connection
                plt.plot(*np.c_[a, b],
                         c=plt.cm.tab10(mesh.fiber.iloc[i] % 10))
            if mesh.iloc[i].z < mesh.iloc[k].z:
                # Draw inter-fiber connection
                plt.plot(*np.c_[a, c], '--ok',
                         lw=1, mfc='none', ms=3, alpha=0.2)
            if mesh.iloc[i].fiber == mesh.iloc[k].fiber:
                # Draw fiber end nodes
                plt.plot(*np.c_[a, c], '+k', ms=3, alpha=0.2)
    # Set drawing box dimensions
    ax.set_xlim(-0.5 * mesh.attrs["size"], 0.5 * mesh.attrs["size"])
    ax.set_ylim(-0.5 * mesh.attrs["size"], 0.5 * mesh.attrs["size"])
    plt.show()
